// rearranged LD3320 sample code, for tests on both asr and mp3 playing by ken woo 20220614.v.0.2.
// 20220618 v.0.3, support esp-m3 which disabled tx/rx. and with some structural modifies.
// 20220624 v.0.4, example, now uses station mode and auto connect and supports multiple files upload. add mp3 cont mode.
// 20220626 v.0.5. debug.
// 20220629 v.0.6, finalized.
// 20220629 v.0.6, example01.
//
// this example usage: 1) prepare enough filesystem space. 2) write flash by earse flash to all contents.
// 3) pc wl-connects to this ap mode and uploads mp3 files by http://192.168.x.x/upload. 4) asr-ing.
// 5) afterwards, if want to write flash, must earse only sketch otherwise the mp3 files would gone.
//
// note1: if enabled MYDB, be care of occasions of violations of playing MP3 and accessing file system via MYDB.
// note2: set ap_mode_or_sta_mode=1 to be sta, 0 to be ap.
// note3: set sta_do_the_connect=1 to demand connecting to ap.
// note4: the remote service {"motor", 8080, "tcp"} is used.
//
// 20220629 v.0.6-1, example01. fixed a bug here and added some note. add other used libs at this pack.(need all in folder words_new)
// 20220709 v.0.8-2, example01. adjust-sensitivity enabled and other minor modified. note that when write flash might cause
//                                 voilation with isr and wdt-reset, disable isr while write could be alright.




#include "LD3320.h"
#include"tables.h"

#define POLL_DEV_PRD            8000 // every 8 seconds to poll device.




//AAAAAAAAAAA/////////////////////////////////////////////////////////////////////////////////////////////////// the IoTs
#define MY_VERSION_NUMBER "[20220629.0.6]"

#include <ESP8266WiFi.h>
#include <ESP8266WebServer.h>
#include <ESP8266HTTPUpdateServer.h>
#include <ESP8266httpUpdate.h>
#include <ESP8266mDNS.h>
#include <ESP8266WiFiMulti.h>
#include <WiFiUdp.h>


#define MYDB 1
#if MYDB
    #include"Esp8266Vars2File.h" // if it is not used, must set ap_mode_or_sta_mode either AP(0) or STA(1).
    // for this lib being supported, the ap/sta mode is kept even poweroff.
    // and even which mode set, every 10 times of power up, the ap mode will be used once. reboot to recover.
#endif // MYDB


#define MYDEBUG 1
#if MYDEBUG
    #define MY_DBG(x) x
#else
    #define MY_DBG(x)
#endif


// -------------------------- for part of hostname --- note: this config will be a service info as well -----------------------
#define LSERVICE_1 "voice-recog" // name


// -------------------------- for default connection names ------------------------------
// i.e., http://hostname.local/
#define HOSTNAME_BASE "esp8266-iot" // for local used.
#define HOSTNAME ((String(HOSTNAME_BASE)+"-"+LSERVICE_1).c_str()) // for local used.


// -------------------------- other settings ------------------------------
// all the TMO values is restricted to no more than 35 minutes; see the Timeout() for detail. however,
// cascading could overcome.
#define TMO_IOT_LOOP_PERIOD 75 // IoT polling at each 75ms
#define TMO_STA_CONN_HEALTH_PERIOD 6000 // each 6 seconds to check the connection health by STA(while ping is alive)
#define TMO_MDNS_Q_PERIOD 30000 // each 30 seconds to query services

// [Multicast]
#define MULTICAST_ADDRESS "239.255.255.250"
#define MULTICAST_PORT "8888"
#define MULTICAST_BUFF_SIZE 15 // default is 16 in total. only (limited)for receiver.
#define MULTICAST_TR 1 // transmitting mode
#define MULTICAST_RV 2 // receiving mode


// -------------------------- main-control variables ------------------------------

const char *const WIFI_CRED[][2]={ // wifi credential
    {"ESP_TESTAP", "87654321"}, // the first one will be used which is in AP mode.
    {"Office", "123456"},
    {"Xperia X Performance_bc10", "e848acbcfd98"},
    {"etc", "etc"},
};

typedef struct smDnsRecord{ // for mdns
    const char *service;
    const unsigned port;
    const char *proto;
} smDnsRecord;

const smDnsRecord mdns_sskn[]={ // for mdns. the known services this device needs from other devices.
    {"http", 80, "tcp"},
    {"https", 443, "tcp"},
    {"test", 81, "tcp"},
    {"motor", 8080, "tcp"},
#if STA_MODE
    {"multicast", 8888, "udp"},
#endif // STA_MODE
};

bool mdns_enabled=false; // set true when connected. could be disabled once aimed and got the ip.

// for demanding STA to connect AP or not, default is yes, controlled by user.
unsigned sta_do_the_connect=0;

// for demanding this local device broadcast or not. it is supposed to switched by a http request.
unsigned multicast_is_broadcasting=0; // 0, disable; 1/MULTICAST_TR, as the transmitter; 2/MULTICAST_RV, as the receiver. and running.

// for where multicasted this device got from of the string. however, since the role can not both at the same time,
// the multicast receiver could still use this str to broadcast.
String multicast_got_or_put_str="";

// for http GET returned
String http_got_str="";

// for connected remote ap ip, in STA mode
char remote_ap_ip[16] = {0};

// remote http server ip, for doing http requests. the auto-http requests is on only if this var assigned.
char http_request_ip[16] = {0};

unsigned is_station_mode_established=0; // change bool to unsigned because elsewhere InTheIram used.

unsigned wifi_ap_credential_idx=0; // for setting up the ap credential index in WIFI_CRED. default is 0.

String ap_credential_custom_name, ap_credential_custom_pw, sta_credential_custom_name, sta_credential_custom_pw; // runtime specified.

ESP8266WebServer *server_p;

ESP8266HTTPUpdateServer httpUpdater; // webupdater, will be http://HOSTNAME.local/update.

ESP8266WiFiMulti *WiFiMulti=0; // for sta mode.

String webupd_addr=""; // the target url for this device processing remote update.

unsigned force_update=0; // force remote update to start immediately.

String uri_list=""; // for listing the uris this device provided.

String upload_abort_filename; // for storing the file name which broken when upload.

unsigned ap_mode_or_sta_mode=0; // be ap or sta, default is ap. could be changed the role by http request /tosta or toap.


// -------------------------- ap mode ------------------------------

bool AP_enter_ap_mode(){ // use WIFI_CRED. default is wifi_ap_credential_idx=0. idx can be changed by http request of /getap and /setap.
    if (ap_credential_custom_name!=""){
        WiFi.softAP(ap_credential_custom_name.c_str(), ap_credential_custom_pw.c_str());
        IPAddress myIP = WiFi.softAPIP();
        MY_DBG(Serial.print("AP IP address: "));
        MY_DBG(Serial.println(myIP));
        return true;
    }
    else if (wifi_ap_credential_idx<sizeof(WIFI_CRED)/sizeof(WIFI_CRED[0])){
        WiFi.softAP(WIFI_CRED[wifi_ap_credential_idx][0], WIFI_CRED[wifi_ap_credential_idx][1]);
        IPAddress myIP = WiFi.softAPIP();
        MY_DBG(Serial.print("AP IP address: "));
        MY_DBG(Serial.println(myIP));
        return true;
    }
    return false;
}


// -------------------------- station mode ------------------------------

bool STA_enter_station_mode(){ // 12 seconds waiting
    unsigned char u=0;

    // Connect to WiFi network
    MY_DBG(Serial.println(""));
    MY_DBG(Serial.println("[-= Enter Station Mode =-]"));
    MY_DBG(Serial.print("WiFi Connecting..."));

    // correct order
    WiFi.disconnect(); // since cases could WiFi reconnect.
    WiFi.mode(WIFI_OFF); // come up with disconnect, not sure ok or not.
    WiFi.mode(WIFI_STA);
    WiFi.begin();
    WiFi.hostname(HOSTNAME); // set after WiFi.begin().

    if (!WiFiMulti){
        WiFiMulti=new ESP8266WiFiMulti;
        if (!WiFiMulti) return false;

        if (sta_credential_custom_name!="")
            WiFiMulti->addAP(sta_credential_custom_name.c_str(), sta_credential_custom_pw.c_str());

        for (int i=0; i<(sizeof(WIFI_CRED)/sizeof(WIFI_CRED[0])); i++)
            WiFiMulti->addAP(WIFI_CRED[i][0], WIFI_CRED[i][1]);
    }

    while (WiFiMulti->run() != WL_CONNECTED) {
        delay(200);
        MY_DBG(Serial.print("."));
        if (u++>60){ // 12 seconds, might not enough
            MY_DBG(Serial.println(" WiFi connection failed..."));
            return false;
        }
    }
    MY_DBG(Serial.println("WiFi connected"));

    // Print the remote MAC address
    uint8_t macAddr[6];
    WiFi.macAddress(macAddr);
    MY_DBG(Serial.printf("Remote MAC adr: %02X:%02X:%02X:%02X:%02X:%02X\r\n",\
        macAddr[0], macAddr[1], macAddr[2], macAddr[3], macAddr[4], macAddr[5]));

    ip_info info;
    wifi_get_ip_info(STATION_IF, &info);
    MY_DBG(Serial.printf("IP ADDR: " IPSTR "\r\n", IP2STR(&info.ip)));
    MY_DBG(Serial.printf("GATEWAY: " IPSTR "\r\n", IP2STR(&info.gw)));
    MY_DBG(Serial.printf("NETMASK: " IPSTR "\r\n", IP2STR(&info.netmask)));

    // resolve the remote ip
    uint32 x=*(uint32*)(&info.ip)&*(uint32*)(&info.netmask);
    ip_addr y=*(ip_addr*)(&x);
    snprintf(remote_ap_ip, 16, "%d.%d.%d.%d",\
        ip4_addr1(&y)? ip4_addr1(&y): 1,\
        ip4_addr2(&y)? ip4_addr2(&y): 1,\
        ip4_addr3(&y)? ip4_addr3(&y): 1,\
        ip4_addr4(&y)? ip4_addr4(&y): 1
        );
    MY_DBG(Serial.print(F("Remote AP IP: ")));
    MY_DBG(Serial.println(remote_ap_ip));

    MY_DBG(Serial.printf("Default local hostname: %s\r\n", WiFi.hostname().c_str()));
    MY_DBG(Serial.print("DNS #1, #2 IP: "));
    MY_DBG(WiFi.dnsIP().printTo(Serial));
    MY_DBG(Serial.print(", "));
    MY_DBG(WiFi.dnsIP(1).printTo(Serial));
    MY_DBG(Serial.println());
    MY_DBG(Serial.printf("RSSI: %d dBm\r\n", WiFi.RSSI()));
    MY_DBG(WiFi.printDiag(Serial));
    
    // Print the local IP address
    MY_DBG(Serial.println(""));
    MY_DBG(Serial.print("IP is assigned: "));
    MY_DBG(Serial.println(WiFi.localIP()));
    return true;
}

void STA_drop_station_mode(){
    WiFi.disconnect();
    WiFi.mode(WIFI_OFF); // no this line will wdt-reset
}


// -------------------------- http server ------------------------------

ESP8266WebServer* HTTPH_create_http_server(){ // ESP8266WebServer & MDNS & Updater facilities
    // Create an instance of the server
    // specify the 80 port to listen on as an argument
    ESP8266WebServer *a=new ESP8266WebServer(80);

    // setup the mdns.
    if (!MDNS.begin(HOSTNAME)) { // Start the mDNS responder
        MY_DBG(Serial.println("Error setting up MDNS responder!"));
    }
    MY_DBG(Serial.println("mDNS responder started"));

    // setup the updater; which is http://hostname/update.
    httpUpdater.setup(a);

    // Start the http server
    a->begin();

    // config mdns.
    MDNS.addService(LSERVICE_1, "tcp", 80);

    MY_DBG(Serial.println(String("UPDATE page is http://")+HOSTNAME+".local:80/update"));
    MY_DBG(Serial.println("HTTP Server started"));
    return a;
}

void remove_the_http_server(){
    if (server_p) delete server_p;
    server_p=0;
}


// -------------------------- http client ------------------------------

bool HTTPC_http_client_connect(const char *url){
    WiFiClient client;
    HTTPClient http; // must be declared after WiFiClient for correct destruction order, because used by http.begin(client,...)
    bool returned_state=false; // true only if got data.

    // configure server and url
    http.begin(client, url);

    // start connection and send HTTP header
    int httpCode = http.GET();
    if (httpCode > 0) {
        // HTTP header has been send and Server response header has been handled

        // file found at server
        if (httpCode == HTTP_CODE_OK) {

            // get lenght of document (is -1 when Server sends no Content-Length header)
            int len = http.getSize();

            #if 0 // with API

                http_got_str=http.getString();
                MY_DBG(Serial.println(http_got_str));

            #else // or "by hand"

                // create buffer for read
                uint8_t buff[128] = {0};
                http_got_str="";

                // get tcp stream
                WiFiClient * stream = &client;

                // read all data from server
                while (http.connected() && (len > 0 || len == -1)) {
                    // read up to 128 byte
                    int c = stream->readBytes(buff, std::min((size_t)len, sizeof(buff)-1));
                    if (!c) {
                        MY_DBG(Serial.println("[read timeout]"));
                        http_got_str="";
                        returned_state=false;
                        break; // must have or it will stuck here! and then if stuck at GET(), set timeout
                    }

                    buff[c]=0;
                    http_got_str+=(char*)buff;

                    if (len > 0) len -= c;
                    returned_state=true;
                }

            #endif // or "by hand"

        }
        else http_got_str="";
    }
    else {
        http_got_str="";
        MY_DBG(Serial.printf("[HTTP GET... failed, error: %s]\r\n", http.errorToString(httpCode).c_str()));
    }
    http.end();
    return returned_state;
}


// -------------------------- mdns ------------------------------

unsigned mdns_sskn_len=sizeof(mdns_sskn)/sizeof(smDnsRecord);

// the count is to pause the auto query for explicitly manual query.
unsigned mDnsQuery_count=0; // the manual query result on how many same service devices were found.

void mDnsQuery(){ // unable to query the provided itselves. especially note that which can be queried should use (USE_MDNS_FULL)
    if (mDnsQuery_count) return;
    MY_DBG(printf("Sending mDNS query\r\n"));
    for (int i=0; i<mdns_sskn_len; i++){
        int n = MDNS.queryService(mdns_sskn[i].service, mdns_sskn[i].proto); // Send out query for those tcp services
        if (!n) MY_DBG(printf("no service [%s] found.\r\n", mdns_sskn[i].service));
        else {
            MY_DBG(printf("service(s) found\r\n"));
            for (int j = 0; j < n; j++)
                MY_DBG(printf("[%s] %d: %s (%s: %d)\r\n",\
                    mdns_sskn[i].service, j+1, MDNS.hostname(j).c_str(), MDNS.IP(j).toString().c_str(), MDNS.port(j)));
        }
    }
    MY_DBG(printf("mDNS query done\r\n"));
}

unsigned mDnsQueryStart(const char *serv_name, const char *serv_protocol){ // start/cont/end are bundled use together.
    // return the number of devices found.
    return mDnsQuery_count=MDNS.queryService(serv_name, serv_protocol);
}

bool mDnsQueryCont(unsigned idx, IPAddress &ip, unsigned &port, String &hostname){ // start/cont/end are bundled use together.
    // use idx to get the found. idx is from 0 to mDnsQuery_count-1.
    if (idx>=mDnsQuery_count) return false;
    hostname=MDNS.hostname(idx);
    ip=MDNS.IP(idx);
    port=MDNS.port(idx);
    return true;
}

void mDnsQueryEnd(){ // start/cont/end are bundled use together.
    mDnsQuery_count=0;
}


// -------------------------- multicast wrapped; multicast is only for sta mode ------------------------------

WiFiUDP Udp;
IPAddress addressMulticast=0;
IPAddress ifaddrMulticast=0;
unsigned portMulticast=0;
char multicast_buffer[MULTICAST_BUFF_SIZE+1]; // default is 16 in total. only (limited)for receiver.

void beginMulticastTransmitter(const char *mcast_address=MULTICAST_ADDRESS, const char *port=MULTICAST_PORT){
    // the timing and contents and tr or recv all are controlled by client code.
    if (ap_mode_or_sta_mode){ // sta mode.
        ifaddrMulticast=WiFi.localIP();
        addressMulticast=ipaddr_addr(mcast_address);
        portMulticast=String(port).toInt();
    }
}

void beginMulticastReceiver(const char *mcast_address=MULTICAST_ADDRESS, const char *port=MULTICAST_PORT){
    // the timing and contents and tr or recv all are controlled by client code.
    if (ap_mode_or_sta_mode){ // sta mode.
        ifaddrMulticast=WiFi.localIP();
        addressMulticast=ipaddr_addr(mcast_address);
        portMulticast=String(port).toInt();
        Udp.beginMulticast(ifaddrMulticast, addressMulticast, portMulticast);
    }
}

void transmitMulticast(){ // the timing and contents and tr or recv all are controlled by client code.
    if (ap_mode_or_sta_mode && (multicast_is_broadcasting==MULTICAST_TR)){
        Udp.beginPacketMulticast(addressMulticast, portMulticast, ifaddrMulticast);
        Udp.write(multicast_got_or_put_str.c_str());
        Udp.endPacket();
    }
}

bool receiveMulticast(){ // the timing and contents and tr or recv all are controlled by client code.
    if (ap_mode_or_sta_mode && (multicast_is_broadcasting==MULTICAST_RV)){
        int packetLength=Udp.parsePacket();
        if (packetLength>0){
            if (packetLength>MULTICAST_BUFF_SIZE) packetLength=MULTICAST_BUFF_SIZE;
            Udp.read(multicast_buffer, packetLength);
            multicast_buffer[packetLength]=0;
            multicast_got_or_put_str=multicast_buffer;
            ///// use IPAddress destinationIP() to identify it is for multicast address.
            ///// use IPAddress remoteIP() to identify the remote address.
            ///// use uint16_t remotePort() to identify the remote port.
            return true;
        }
    }
    return false;
}

void stopMulticast(){ // if stop, multicast_is_broadcasting must reassign if reuse.
    // the timing and contents and tr or recv all are controlled by client code.

    if (ap_mode_or_sta_mode){
        multicast_is_broadcasting=0;
        Udp.stop(); // especially note that the receiver is of concerns, so better to stop if not used.
    }
}


// -------------------------- Remote updater ------------------------------
// the device can not update itself because this updater has yet not found any exit point.

void gbl_update_started() {
  MY_DBG(Serial.println("CALLBACK:  HTTP update process started"));
}

void gbl_update_finished() {
  MY_DBG(Serial.println("CALLBACK:  HTTP update process finished"));
}

void gbl_update_progress(int cur, int total) {
  MY_DBG(Serial.printf("CALLBACK:  HTTP update process at %d of %d bytes...\n", cur, total));
}

void gbl_update_error(int err) {
  MY_DBG(Serial.printf("CALLBACK:  HTTP update fatal error code %d\n", err));
}

void gbl_updater(const char *url) {
    // wait for WiFi connection

    WiFiClient client;

    // The line below is optional. It can be used to blink the LED on the board during flashing
    // The LED will be on during download of one buffer of data from the network. The LED will
    // be off during writing that buffer to flash
    // On a good connection the LED should flash regularly. On a bad connection the LED will be
    // on much longer than it will be off. Other pins than LED_BUILTIN may be used. The second
    // value is used to put the LED on. If the LED is on with HIGH, that value should be passed
    // ESPhttpUpdate.setLedPin(LED_BUILTIN, LOW);

    // Add optional callback notifiers
    ESPhttpUpdate.onStart(gbl_update_started);
    ESPhttpUpdate.onEnd(gbl_update_finished);
    ESPhttpUpdate.onProgress(gbl_update_progress);
    ESPhttpUpdate.onError(gbl_update_error);

    t_httpUpdate_return ret = ESPhttpUpdate.update(client, url);
    // Or:
    // t_httpUpdate_return ret = ESPhttpUpdate.update(client, "server", 80, "file.bin");

    switch (ret) {
        case HTTP_UPDATE_FAILED:
            MY_DBG(Serial.printf("HTTP_UPDATE_FAILD Error (%d): %s\r\n", ESPhttpUpdate.getLastError(), ESPhttpUpdate.getLastErrorString().c_str()));
            break;
        case HTTP_UPDATE_NO_UPDATES:
            MY_DBG(Serial.println("HTTP_UPDATE_NO_UPDATES"));
            break;
        case HTTP_UPDATE_OK:
            MY_DBG(Serial.println("HTTP_UPDATE_OK"));
            break;
    }
}


// -------------------------- for files upload, uri: /upload. and the list dir/del file facilities. ---------------------------
String get_content_type(String filename){
    if (filename.endsWith(".htm")) return "text/html";
    else if (filename.endsWith(".html")) return "text/html";
    else if (filename.endsWith(".css")) return "text/css";
    else if (filename.endsWith(".js")) return "application/javascript";
    else if (filename.endsWith(".png")) return "image/png";
    else if (filename.endsWith(".gif")) return "image/gif";
    else if (filename.endsWith(".jpg")) return "image/jpeg";
    else if (filename.endsWith(".ico")) return "image/x-icon";
    else if (filename.endsWith(".xml")) return "text/xml";
    else if (filename.endsWith(".pdf")) return "application/x-pdf";
    else if (filename.endsWith(".zip")) return "application/x-zip";
    else if (filename.endsWith(".gz")) return "application/x-gzip";
    else if (filename.endsWith(".bin")) return "application/octet-stream";
    else if (filename.endsWith(".mp3")) return "application/octet-stream";
    return "text/plain";
}

bool page_handleFileRead(String path){ // send the file to the client if it exists
    LittleFS.begin();
    if (path.endsWith("/")) path += "index.html"; // If a folder is requested, send the index file
    String pathWithGz = path + ".gz";
    if (LittleFS.exists(pathWithGz) || LittleFS.exists(path)) { // If the file exists, either as a compressed archive, or normal
        if (LittleFS.exists(pathWithGz)) path = pathWithGz; // If there's a compressed version available, then use the compressed version
        File file = LittleFS.open(path, "r"); // Open the file
        size_t sent = server_p->streamFile(file, get_content_type(path)); // Send it to the client
        file.close(); // Close the file
        MY_DBG(Serial.println(String("Sent the file done: ") + path));
        LittleFS.end();
        return true;
    }
    MY_DBG(Serial.println(String("File Not Found: ") + path));
    // If the file doesn't exist, return false and direct back to notify false page.

    LittleFS.end();
    return false;
}

void page_handleFileUpload_cb(){

    extern String upload_abort_filename;
    upload_abort_filename="";

    static File fsu_file;
    HTTPUpload &upload = server_p->upload();
    String filename = upload.filename;
    switch (upload.status){
        case UPLOAD_FILE_START:
            LittleFS.begin();
            if (!filename.startsWith("/")) filename = "/"+filename;
            MY_DBG(printf("handle the file upload: %s\r\n", filename.c_str()));
            fsu_file = LittleFS.open(filename, "w"); // Open the file for writing in littlefs (create if it doesn't exist)
            break;
        case UPLOAD_FILE_WRITE:
            if (fsu_file)
                fsu_file.write(upload.buf, upload.currentSize); // Write the received bytes to the file
            break;
        case UPLOAD_FILE_END:
            if (fsu_file) { // If the file was successfully created
                fsu_file.close(); // Close the file
                MY_DBG(printf("done the file, total size: %u\r\n", upload.totalSize));
                server_p->sendHeader("Location", "/success.html"); // Redirect the client to the success page
                server_p->send(303);
            }
            else {
                server_p->send(500, "text/plain", "500: could not create file");
                MY_DBG(printf("error, could not create file: %s\r\n", filename.c_str()));
            }
            LittleFS.end();
            break;
        default:
            LittleFS.end();
            MY_DBG(printf("upload error. code: %d, fn: %s\r\n", upload.status, filename.c_str()));
            upload_abort_filename=filename;
            break;
    }
}

void Ls(String *s=0, const char *path="/"){
    LittleFS.begin();

    Dir dir = LittleFS.openDir(path);

    if (s){
        (*s)+="list directory \"";
        (*s)+=path;
        (*s)+="\":\r\n--------\r\n";
    }
    else MY_DBG(printf("list directory \"%s\":\r\n--------\r\n", path));
    while (dir.next()){
        if (s){
            (*s)+=dir.fileName();
            (*s)+="    ";
            (*s)+=(dir.isDirectory()? "<DIR>": (String(dir.fileSize())+" bytes").c_str());
            (*s)+="\r\n";
        }
        else
            MY_DBG(printf("%s    %s\r\n",\
                dir.fileName().c_str(), dir.isDirectory()? "<DIR>": (String(dir.fileSize())+" bytes").c_str()));
    }

    FSInfo fs_info;
    if (LittleFS.info(fs_info)){
        if (s){
            (*s)+="\r\nentire space ";
            (*s)+=fs_info.totalBytes;
            (*s)+=" bytes\r\n";
            (*s)+=fs_info.usedBytes;
            (*s)+=" bytes used\r\nfree space ";
            (*s)+=fs_info.totalBytes-fs_info.usedBytes;
            (*s)+=" bytes\r\n--------\r\n";
        }
        else MY_DBG(printf("\r\nentire space %d bytes\r\n%d bytes used\r\nfree space %d bytes\r\n--------\r\n",\
            fs_info.totalBytes, fs_info.usedBytes, fs_info.totalBytes-fs_info.usedBytes));
    }

    LittleFS.end();
}


// -------------------------- other functions ------------------------------
bool Timeout(unsigned ms, unsigned &store_start_time_us, bool one_shot=false, bool one_shot_lasting=false){
    if (one_shot && !store_start_time_us) return one_shot_lasting;
    if (int(micros())-int(ms*1000+store_start_time_us)>=0){ ///// this is ok
        store_start_time_us=one_shot? 0: micros();
        return true;
    }
    return false;
}


// -------------------------- main controls ------------------------------

void the_http_request(){

    extern void my_http_request(); // all the http requests from this device could be controlled in this function.
    if (*http_request_ip) my_http_request(); // called only if http_request_ip assigned.

}

void the_http_server(){

    server_p=HTTPH_create_http_server();

    server_p->on("/", [](){
        server_p->send(200, "text/html", ap_mode_or_sta_mode? "<h1>WiFi STATION</h1>": "<h1>WiFi AP</h1>");
    });

    server_p->on("/version", []() {server_p->send(200, "text/plain", MY_VERSION_NUMBER);});

    server_p->on("/reset", []() {
        server_p->send(200, "text/plain", "now reset...");
        delay(1000);
        while (1);
    });

    server_p->on("/s", []() {server_p->send(200, "text/plain", String(random(10)+1).c_str());});

    server_p->on("/getap", [](){
        String a("choose the number:\r\n");
        for (int i=0; i<(sizeof(WIFI_CRED)/sizeof(WIFI_CRED[0])); i++){
            a+=String(i+1)+". ["+WIFI_CRED[i][0]+", "+/*WIFI_CRED[i][1]*/"***"+"]\r\n";
        }
        a=a+"currently is "+(wifi_ap_credential_idx+1)+". ["+WIFI_CRED[wifi_ap_credential_idx][0]+"]\r\n";
        server_p->send(200, "text/plain", a.c_str());
    });

    server_p->on("/setap", [](){ // ?n=<i>.
        String n=server_p->arg("n");
        int nn=n.toInt()-1;
        if (nn>=0 && nn<sizeof(WIFI_CRED)/sizeof(WIFI_CRED[0])) wifi_ap_credential_idx=nn;
        else wifi_ap_credential_idx=0;

        #if MYDB
            cDataBase("/settings").ReplaceWordsEx("wifi_ap_credential_idx", 0, wifi_ap_credential_idx);
        #endif // MYDB

        server_p->send(200, "text/plain", (String(WIFI_CRED[wifi_ap_credential_idx][0])+" chosen.").c_str());
    });

    server_p->on("/toap", [](){
        server_p->send(200, "text/plain", "done.");
        ap_mode_or_sta_mode=0;

        #if MYDB
            cDataBase("/settings").ReplaceWordsEx("ap_mode_or_sta_mode", 0, ap_mode_or_sta_mode);
        #endif // MYDB

        delay(1000);
        while (1); // reset to change role.
    });

    server_p->on("/tosta", [](){
        server_p->send(200, "text/plain", "done.");
        ap_mode_or_sta_mode=1;

        #if MYDB
            cDataBase("/settings").ReplaceWordsEx("ap_mode_or_sta_mode", 0, ap_mode_or_sta_mode);
        #endif // MYDB

        delay(1000);
        while (1); // reset to change role.
    });

    server_p->on("/credential", [](){ // ?r=<the role is ap or sta>&n=<name>&p=<password>.
        server_p->send(200, "text/plain", "done.");
        String role=server_p->arg("r");
        String cname=server_p->arg("n");
        String pw=server_p->arg("p");

        #if MYDB
            unsigned n_len=cname.length()+1, p_len=pw.length()+1;
            if (role=="ap"){
                cDataBase("/settings").ReplaceBytesEx("ap_credential_custom_name", 0, cname.c_str(), n_len, n_len);
                cDataBase("/settings").ReplaceBytesEx("ap_credential_custom_pw", 0, pw.c_str(), p_len, p_len);
            }
            else if (role=="sta"){
                cDataBase("/settings").ReplaceBytesEx("sta_credential_custom_name", 0, cname.c_str(), n_len, n_len);
                cDataBase("/settings").ReplaceBytesEx("sta_credential_custom_pw", 0, pw.c_str(), p_len, p_len);
            }
        #endif // MYDB

    });

    // assign the update website url to this device(it is intended for this local device connecting to)
    server_p->on("/webupdinfo", []() { // http://<ip address>/webupdinfo?addr=<remote url; e.g., http://192.168.1.1/a.bin>.
        String addr=server_p->arg("addr");
        if (addr=="") server_p->send(404, "text/plain", "0. page not found.");
        else {
            webupd_addr=addr;
            server_p->send(200, "text/plain", ("1. accepted: "+webupd_addr).c_str());
        }
    });

    // force global update immediately
    server_p->on("/forceupdate", []() {
        force_update=1;
        server_p->send(200, "text/plain", String("1. accepted. on 10 seconds to start updating...").c_str());
    });

    server_p->on("/listcmds", []() { // list all commands.
        server_p->send(200, "text/plain", uri_list.c_str());
    });

    // list directories.
    server_p->on("/ls", []() {
        String a;
        Ls(&a);
        server_p->send(200, "text/plain", a.c_str());
    });

    // delete a file.
    server_p->on("/del", []() { // ?n=<filename>.
        String n=server_p->arg("n");
        LittleFS.begin();
        LittleFS.remove(n);
        LittleFS.end();
        server_p->send(200, "text/plain", "done");
    });

    server_p->on("/success.html", HTTP_GET, []() {

        extern String upload_abort_filename;

        server_p->send(
            200, "text/html",\
            (upload_abort_filename=="")\
                ? "<!DOCTYPE HTML>\r\n<html>\r\nsuccess\r\n</html>"\
                : (String("<!DOCTYPE HTML>\r\n<html>\r\nabort at file: ")+\
                    upload_abort_filename+\
                    String("\r\n</html>")).c_str()\
        );
    });

    // upload a file to this device. POST.
    server_p->on("/upload", HTTP_POST, []() {
            server_p->send(200);
        },
        page_handleFileUpload_cb
    );

    // upload a file to this device. GET.
    server_p->on("/upload", HTTP_GET, []() {
        if (!page_handleFileRead("/upload.html")){
            server_p->send(200, "text/html", "<!DOCTYPE HTML>\r\n<html>\r\n\
                <form method=\"post\" enctype=\"multipart/form-data\">\
                <input type=\"file\" name=\"name\" multiple>\
                <input class=\"button\" type=\"submit\" value=\"Upload\"></form>\r\n\
                </html>");
        }
    });

    // use uri_list+=str; to expand uri list.
    uri_list=\
        "/listcmds\r\n"\
        "/\r\n"\
        "/version\r\n"\
        "/reset\r\n"\
        "/s\r\n"\
        "/update (by selecting an update file from pc and upload to update)\r\n"\
        "/webupdinfo?addr=<remote url; e.g., http://192.168.1.1/a.bin>\r\n"\
        "/forceupdate\r\n"\
        "/getap\r\n"\
        "/setap?n=<i>\r\n"\
        "/credential?r=<the role is ap or sta>&n=<name>&p=<password>\r\n"\
        "/toap\r\n"\
        "/tosta\r\n"\
        "/ls (list root files and directories)\r\n"\
        "/del?n=<filename; delete this file>\r\n"\
        "/upload\r\n";
}

void setup_IoT(){
    if (ap_mode_or_sta_mode) MY_DBG(Serial.println("station mode entered."));
    else {
        MY_DBG(Serial.println("Configuring access point..."));
        AP_enter_ap_mode();
        mdns_enabled=true;

        the_http_server(); // setup the http server
    }
}


// ------------------------------------------new added for my application, of the IoTs.
bool poll_device=true;
const char *device_name="motor"; // protocol tcp, port 8080.
String target_dev_ip, target_dev_last_sent;

unsigned successive_cmd_issue=0;
int successive_cmd_status=0;

void loop_Remote(){
    static unsigned wtmp=0;
    if (poll_device){ // device still not found and needs to poll.
        static unsigned tmoz=micros();
        if (!Timeout(POLL_DEV_PRD, tmoz)) return; // to poll device

        IPAddress ip;
        unsigned port;
        String a;
        int i=mDnsQueryStart(device_name, "tcp");
        for (int j=0; j<i; j++){
            if (mDnsQueryCont(j, ip, port, a)){
                target_dev_ip=ip.toString();
                poll_device=false;
                break;
            }
        }
    }
    else {
        static unsigned xtmp=1;
        switch (successive_cmd_issue){
            case 2: // open
                HTTPC_http_client_connect(("http://"+target_dev_ip+"/params/gate/open").c_str());
                target_dev_last_sent=http_got_str;
                break;
            case 3: // close
                HTTPC_http_client_connect(("http://"+target_dev_ip+"/params/gate/close").c_str());
                target_dev_last_sent=http_got_str;
                break;
            case 5: // probe
                HTTPC_http_client_connect(("http://"+target_dev_ip+"/params/mtr/probe").c_str());
                target_dev_last_sent=http_got_str;
                break;
            case 6: // syncon
                HTTPC_http_client_connect(("http://"+target_dev_ip+"/params/mtr/syncon").c_str());
                delay(800);
                HTTPC_http_client_connect(("http://"+target_dev_ip+"/params/mtr/on").c_str());
                target_dev_last_sent=http_got_str;
                break;
            case 7: // small-open
                HTTPC_http_client_connect(("http://"+target_dev_ip+"/params/gate/efficiency").c_str());
                target_dev_last_sent=http_got_str;
                break;
            case 8: // large-open
                HTTPC_http_client_connect(("http://"+target_dev_ip+"/params/gate/performance").c_str());
                target_dev_last_sent=http_got_str;
                break;
            case 9: // inc-open
                if (xtmp){
                    xtmp=0;
                    HTTPC_http_client_connect(("http://"+target_dev_ip+"/params/mtr/syncoff").c_str());
                    delay(800);
                }
                HTTPC_http_client_connect(("http://"+target_dev_ip+"/t?t=3").c_str());
                target_dev_last_sent=http_got_str;
                break;
            case 10: // inc-close
                if (xtmp){
                    xtmp=0;
                    HTTPC_http_client_connect(("http://"+target_dev_ip+"/params/mtr/syncoff").c_str());
                    delay(800);
                }
                HTTPC_http_client_connect(("http://"+target_dev_ip+"/t?t=4").c_str());
                target_dev_last_sent=http_got_str;
                break;
            case 11: // loose-close
                HTTPC_http_client_connect(("http://"+target_dev_ip+"/params/gate/looseclose").c_str());
                target_dev_last_sent=http_got_str;
                break;
            default: break;
        }
        successive_cmd_issue=0;
    }
}
// ------------------------------------------new added for my application, of the IoTs.


void loop_IoT(){
    static unsigned tmo1=micros();
    if (!Timeout(TMO_IOT_LOOP_PERIOD, tmo1)) return; // polling at each 75ms

    if (ap_mode_or_sta_mode){ // station mode.
        if (!sta_do_the_connect){
            if (is_station_mode_established){
                is_station_mode_established=0;
                STA_drop_station_mode();
            }
            return; // here will return if no station connection, so latter code would not run.
        }
        static unsigned tmo2=micros(), tmo02=tmo2; // tmo02, after 3 seconds starts to connect.
        if (Timeout(3000, tmo02, true) || Timeout(TMO_STA_CONN_HEALTH_PERIOD, tmo2, true, true)){
            if (!is_station_mode_established){
                mdns_enabled=is_station_mode_established=STA_enter_station_mode(); // it is blocking wait.
                static unsigned fault_retry_cnt=0;
                if (is_station_mode_established){ // run at here since mdns needs the ip or wdt-reset.
                    fault_retry_cnt=0;

                    remove_the_http_server();
                    the_http_server(); // setup the http server

                }
                else { // a fault case that resulting wdt reset and then sta always invalid, so wdt reset again to retry!!!!!
                    tmo2=micros();
                    if (fault_retry_cnt++>25) while (1);
                }
                return;
            }
        }
        else return;


// ------------------------------------------new added for my application, of the IoTs.
        switch (successive_cmd_status){
            case 1: // when demand ap-connection, set status to 1 to polling wifi-conn-status.
                if (is_station_mode_established) successive_cmd_status=2; // on to line.
                break;
            case 3: // searching device phase.
                loop_Remote();
                if (!poll_device) successive_cmd_status=4; // the remote device connected.
                break;
            case 5: // accessing device phase.
                loop_Remote();
                break;
            default: break;
        }
// ------------------------------------------new added for my application, of the IoTs.


    }

    if (mdns_enabled){
        MDNS.update();

        //// static unsigned tmo4=micros();
        //// if (Timeout(TMO_MDNS_Q_PERIOD, tmo4)) mDnsQuery(); // each 30 seconds to query services.

    }

    server_p->handleClient(); // handles http requests from remotes.

    if (force_update && webupd_addr!=""){ // handles update.
        static unsigned tmo3=micros();
        if (force_update==1){
            force_update=2;
            tmo3=micros();
        }
        else if ((force_update==2) && Timeout(10000, tmo3, true)){ // 10 seconds to start
            force_update=0;
            gbl_updater(webupd_addr.c_str());
        }
    }

    the_http_request(); // for local http request to remote
}

void my_http_request(){;}


// -------------------------- main loop ------------------------------

void the_iot_setup(){
    #if MYDB
        unsigned boot_up_times=0;
        unsigned char buff[32];
        cDataBase db("/settings");
        db.Open();
        db.ReadWords(boot_up_times, "boot_up_times");
        db.ReadWords(wifi_ap_credential_idx, "wifi_ap_credential_idx");
        db.ReadWords(ap_mode_or_sta_mode, "ap_mode_or_sta_mode");

        buff[0]=0; db.ReadData(buff, 32, "ap_credential_custom_name", 0); ap_credential_custom_name=(char*)buff;
        buff[0]=0; db.ReadData(buff, 32, "ap_credential_custom_pw", 0); ap_credential_custom_pw=(char*)buff;
        buff[0]=0; db.ReadData(buff, 32, "sta_credential_custom_name", 0); sta_credential_custom_name=(char*)buff;
        buff[0]=0; db.ReadData(buff, 32, "sta_credential_custom_pw", 0); sta_credential_custom_pw=(char*)buff;

        db.Close();

        ++boot_up_times;
        cDataBase("/settings").ReplaceWordsEx("boot_up_times", 0, boot_up_times);
        if (!(boot_up_times%10)){
            cDataBase("/settings").Renew();

            ap_mode_or_sta_mode=0; // force to in ap mode, reboot to recover.
        }
    #endif // MYDB

    setup_IoT();
}
//VVVVVVVVVVV/////////////////////////////////////////////////////////////////////////////////////////////////// the IoTs




//AAAAAAAAAAA/////////////////////////////////////////////////////////////////////////////////////////////////// LD3320

IRAM_ATTR volatile unsigned &HardwareWDT = *((volatile unsigned*)0x60000900);
IRAM_ATTR volatile unsigned &HwTimer1Load = *((volatile unsigned*)0x60000600);
#define disableHardwareWDT  (HardwareWDT&=~1)
#define enableHardwareWDT   (HardwareWDT|=1)

void halt_wdt(){
    static unsigned t1=1;
    if (t1){
        t1=0;
        WiFi.mode(WIFI_OFF);
        WiFi.forceSleepBegin(-1);
        system_soft_wdt_feed();
        system_soft_wdt_stop();
        disableHardwareWDT;
        delay(1500);
        MY_DBG(Serial.printf("\r\nHALT IOT and WDT\r\n"));
    }
}




cLinkedTables *voice_tables;
unsigned process_level=0, process_level_back=0;
unsigned the_debug_data, the_debug_cnt;


unsigned answering_table_idx=1; // for startup.

mytable answering_table={
    "",
    "/startup.mp3",
    "sl hi there sl, my name is mary, nice to meet you sl sl",
    "the y 5 is connected, sl enjoy it sl.",
    "motor device is found, sl and gotten in touch sl.",
    "pardon me sir sl, please say it again sl thank you sl sl.",
    "i will switch to a p mode and reboot sl sl",
    "i will switch to s t a mode and reboot sl sl",
    "command sent successfully sl sl",
    "command sent failed sl sl",
    "/sample1.mp3",
    "/sample2.mp3",
    "\\19216811",
};




// note: the tables structure, voc0(to recog) -> foo0(to invoke) -> voc1 -> foo1, etc.
mytable voc_tbl0={
    "xiao jie",             // 0 小姐
    "pu long gong",         // 1 普攏拱
    "woo own hon",          // 2
    "hoo own own",          // 3

    "en zi si kou pu xu",   // 4 恩茲思扣撲虛
    "pu kou kou",           // 5
    "pu pu kou",            // 6
    "si zi",                // 7
    "zi si",                // 8
    "si zi gi",             // 9
    "sa je",                // 10
    "xiao zi",              // 11
    ""
};

void foo_level0_common(){
    voice_tables->moveToNext(voice_tables);
    voice_tables->moveToNext(voice_tables);
    LD3320::LD_ChangeTable(voice_tables->getT());

    process_level=2;

  MY_DBG(Serial.println("-------------------------> to level 1\r\n"));
  MY_DBG(voice_tables->debug_dumpTable(*(char*(*)[])LD3320::LD_GetTable()));
}
void foo_level0_garbage(unsigned a){
    switch (a){
        case 4: MY_DBG(Serial.println("-------------------------> 4 en zi si kou pu xu\r\n")); break;
        case 5: MY_DBG(Serial.println("-------------------------> 5 pu kou kou\r\n")); break;
        case 6: MY_DBG(Serial.println("-------------------------> 6 pu pu kou\r\n")); break;
        case 7: MY_DBG(Serial.println("-------------------------> 7 si zi\r\n")); break;
        case 8: MY_DBG(Serial.println("-------------------------> 8 zi si\r\n")); break;
        case 9: MY_DBG(Serial.println("-------------------------> 9 si zi gi\r\n")); break;
        case 10: MY_DBG(Serial.println("-------------------------> 10 sa je\r\n")); break;
        case 11: MY_DBG(Serial.println("-------------------------> 11 xiao zi\r\n")); break;
        default: MY_DBG(Serial.println("-------------------------> xxxxxxxxx\r\n")); break;
    }
}
void foo_xiao_jie(unsigned a){
    MY_DBG(Serial.println("-------------------------> 0 小姐\r\n"));


    LD3320::StopASR(); // since it is modified to idle.
    nAsrStatus=LD_ASR_NONE; // since it is modified to idle.
    //LD3320MP3("/sample1.mp3");
    MP3Sentence t("yes sir, what can i help you.", 2);
}
void foo_pu_long_gong(unsigned a){
    foo_level0_common();
    switch (a){
        case 1: MY_DBG(Serial.println("-------------------------> 1 普攏拱\r\n")); break;
        case 2: MY_DBG(Serial.println("-------------------------> 2 普攏拱\r\n")); break;
        case 3: MY_DBG(Serial.println("-------------------------> 3 普攏拱\r\n")); break;
        default: MY_DBG(Serial.println("-------------------------> xxxxxxxxx\r\n")); break;
    }


    LD3320::StopASR();
    nAsrStatus=LD_ASR_NONE;
    //LD3320MP3::D("/sample2.mp3");
    MP3Sentence t("o fxxk you sl, ken woo sl, what is up.", 3);
}

mytable foo_tbl0={
    (mycstr)foo_xiao_jie,                   // "xiao jie",             // 小姐
    (mycstr)foo_pu_long_gong,               // "pu long gong",         // 普攏拱
    (mycstr)foo_pu_long_gong,               // "woo own hon",
    (mycstr)foo_pu_long_gong,               // "hoo own own",
    (mycstr)foo_level0_garbage,             // "en zi si kou pu xu",   // 恩茲思扣撲虛
    (mycstr)foo_level0_garbage,             // "pu kou kou",
    (mycstr)foo_level0_garbage,             // "pu pu kou",
    (mycstr)foo_level0_garbage,             // "si zi",
    (mycstr)foo_level0_garbage,             // "zi si",
    (mycstr)foo_level0_garbage,             // "si zi gi",
    (mycstr)foo_level0_garbage,             // "sa je",
    (mycstr)foo_level0_garbage,             // "xiao zi",
    ""
};


mytable voc_tbl1={
    "fa men kai qi",        // 0 閥門開啟
    "kai qi fa men",        // 1 開啟閥門
    "fa men guan bi",       // 2 閥門關閉
    "guan bi fa men",       // 3 關閉閥門
    "xian zai zhuai su",    // 4 現在轉速
    "fa men zi dong",       // 5 閥門自動
    "fa men suo ding",      // 6 閥門鎖定
    "fa men wei kai",       // 7 閥門微開
    "fa men wei guan",      // 8 閥門微關

    "en zi si kou pu xu",   // 9 恩茲思扣撲虛
    "pu kou kou",           // 10
    "pu pu kou",            // 11
    "si zi",                // 12
    "zi si",                // 13
    "si zi gi",             // 14
    "sa je",                // 15
    "xiao zi",              // 16
    ""
};

void foo_level1_common(){
    voice_tables->moveToPrev(voice_tables);
    voice_tables->moveToPrev(voice_tables);
    LD3320::LD_ChangeTable(voice_tables->getT());

    process_level=1;
    process_level_back=0;

    MY_DBG(Serial.println("-------------------------> to level 0\r\n"));
    MY_DBG(voice_tables->debug_dumpTable(*(char*(*)[])LD3320::LD_GetTable()));
}
void foo_level1_garbage(unsigned a){
    foo_level1_common();
    switch (a) {
        case 9: MY_DBG(Serial.println("-------------------------> 9 en zi si kou pu xu\r\n")); break;
        case 10: MY_DBG(Serial.println("-------------------------> 10 pu kou kou\r\n")); break;
        case 11: MY_DBG(Serial.println("-------------------------> 11 pu pu kou\r\n")); break;
        case 12: MY_DBG(Serial.println("-------------------------> 12 si zi\r\n")); break;
        case 13: MY_DBG(Serial.println("-------------------------> 13 zi si\r\n")); break;
        case 14: MY_DBG(Serial.println("-------------------------> 14 si zi gi\r\n")); break;
        case 15: MY_DBG(Serial.println("-------------------------> 15 sa je\r\n")); break;
        case 16: MY_DBG(Serial.println("-------------------------> 16 xiao zi\r\n")); break;
        default: MY_DBG(Serial.println("-------------------------> xxxxxxxxx\r\n")); break;
    }

    answering_table_idx=5;
}
void foo_fa_men_kai_qi(unsigned a){
    foo_level1_common();
    MY_DBG(Serial.println("-------------------------> 閥門開啟\r\n"));
}
void foo_fa_men_guan_bi(unsigned a){
    foo_level1_common();
    MY_DBG(Serial.println("-------------------------> 閥門關閉\r\n"));
}
void foo_xian_zai_zhuai_su(unsigned a){
    foo_level1_common();
    MY_DBG(Serial.println("-------------------------> 現在轉速\r\n"));
}
void foo_fa_men_zi_dong(unsigned a){
    foo_level1_common();
    MY_DBG(Serial.println("-------------------------> 閥門自動\r\n"));
}
void foo_fa_men_suo_ding(unsigned a){
    foo_level1_common();
    MY_DBG(Serial.println("-------------------------> 閥門鎖定\r\n"));
}
void foo_fa_men_wei_kai(unsigned a){
    foo_level1_common();
    MY_DBG(Serial.println("-------------------------> 閥門微開\r\n"));
}
void foo_fa_men_wei_guan(unsigned a){
    foo_level1_common();
    MY_DBG(Serial.println("-------------------------> 閥門微關\r\n"));
}

mytable foo_tbl1={
    (mycstr)foo_fa_men_kai_qi,              // "fa men kai qi",        // 閥門開啟
    (mycstr)foo_fa_men_kai_qi,              // "kai qi fa men",        // 開啟閥門
    (mycstr)foo_fa_men_guan_bi,             // "fa men guan bi",       // 閥門關閉
    (mycstr)foo_fa_men_guan_bi,             // "guan bi fa men",       // 關閉閥門
    (mycstr)foo_xian_zai_zhuai_su,          // "xian zai zhuai su",    // 現在轉速
    (mycstr)foo_fa_men_zi_dong,             // "fa men zi dong",       // 閥門自動
    (mycstr)foo_fa_men_suo_ding,            // "fa men suo ding",      // 閥門鎖定
    (mycstr)foo_fa_men_wei_kai,             // "fa men wei kai",       // 閥門微開
    (mycstr)foo_fa_men_wei_guan,            // "fa men wei guan",      // 閥門微關
    (mycstr)foo_level1_garbage,
    (mycstr)foo_level1_garbage,
    (mycstr)foo_level1_garbage,
    (mycstr)foo_level1_garbage,
    (mycstr)foo_level1_garbage,
    (mycstr)foo_level1_garbage,
    (mycstr)foo_level1_garbage,
    (mycstr)foo_level1_garbage,
    ""
};


void test1(){
    MP3Sentence("how are you today?", 1);
    MY_DBG({Serial.println("test1"); Serial.println("how are you today?");});
}

unsigned test2(){
    LittleFS.begin();

    static unsigned v=0;
    if (++v>30){v=0; return 0;}

    unsigned q=random(350)+1; /////////////////////////////////// the number of mp3 files.
    Dir dir=LittleFS.openDir("/");
    while (--q && dir.next());
    static char buff[20]; // need static or fault.
    strcpy(buff, dir.fileName().c_str());
    MY_DBG({Serial.print("test2    ");Serial.println(buff);});


    while (LD3320MP3::isOpened() && !LD3320MP3::isDone()) delay(1); // for the previous file closed.
    LD3320MP3::Close(); // twice-close is ok.

    LD3320MP3 t(buff);
    return 1;
}

void test3(){
    LittleFS.begin();

    Dir dir=LittleFS.openDir("/");
    static char buff[128]; // need static or fault.
    String tmp="";

    for (int i=0, q=0; i<20; i++){ /////////////////////////////////// the number of mp3 files 300.
        q=random(15)+2;
        while (--q && dir.next());
        tmp+=dir.fileName().substring(1, dir.fileName().indexOf('.'));
        tmp+="\\";
    }
    strcpy(buff, tmp.c_str());
    MY_DBG({Serial.println("test3"); Serial.println(buff);});
    MP3Sentence((const char*)buff, 2);
}

void test4(){

    static unsigned bt=0;

    MY_DBG(Serial.println("test4"));
    if (++bt==1)
        MP3Sentence::AddSilence("sample2.mp3 _a.mp3 _b.mp3 _c.mp3 sample1.mp3 _d.mp3 _e.mp3 _f.mp3 _g.mp3 startup.mp3", 2);

    else if (bt==2) LD3320MP3("sample1.mp3");

    else if (bt==3) MP3Sentence::AddSilence("sample1.mp3", 3);

    else if (bt==4){
        MY_DBG(Serial.println("hi! how are you sir, would you please give me a favor to open the device?"));
        MP3Sentence t("hi! how are you sir, would you please give me a favor to open the device?", 3);
    }

    else if (bt==5)
        MP3Sentence t("bp bp bp bp bp bp bp bp bp bp bp bp bp bp bp bp bp bp bp bp bp bp bp bp bp bp bp bp", 31);

    else if (bt==6){
        MY_DBG(Serial.println("how 321 are 456 you 987 today"));
        MP3Sentence::Literal("how 321 are 456 you 987 today", 2);
        bt=0;
    }
}

void test5_my_asr_callback(unsigned id){
    // all the voice recognitions identified by ids are via this function.
    MY_DBG(Serial.printf("\r\nthe id is %u, directly response from ASR\r\n", id));
    MY_DBG(Serial.printf("SPEAK: this is the voice recognition device, feel free to enjoy it.\r\n"));
    MP3Sentence t("this is the voice recognition device, sl feel free to enjoy it.", 2);
}




void setup(){
    MY_DBG(Serial.begin(115200));
    MY_DBG(while (!Serial) delay(10));
    MY_DBG(delay(3000));
    MY_DBG(Serial.printf("\r\nStart\r\n"));

    // IoTs
    the_iot_setup();

    // LD3320
    LD3320::LD_Init(); // init.
    nAsrStatus=LD_ASR_NONE; // means regularly doing the asr or mp3s time by time.
    LD3320::LD_AdjustMIX2SPVolume(10); // adjust volume.

    // prepare tables
    voice_tables=new cLinkedTables;
    voice_tables->assignTable(voc_tbl0);
    voice_tables->addTable(foo_tbl0);
    voice_tables->addTable(voc_tbl1);
    voice_tables->addTable(foo_tbl1);
    process_level=1;

    LD3320::LD_ChangeTable(voice_tables->getT());
}

void loop(){

    #if MP3_IN_ISR
        halt_wdt();
    #endif

    // for auto responder, or enquiry.
    if (!answering_table_idx){
        switch (successive_cmd_status){
            case 2: // online
                answering_table_idx=3; // online.
                successive_cmd_status++;
                break;
            case 4: // found the device
                answering_table_idx=4; // got the device.
                successive_cmd_status++;
                break;
            default: break;
        }
    }
    else if (!LD3320MP3::isOpened()){ // play the auto-response sound.
        LD3320::StopASR();
        nAsrStatus=LD_ASR_NONE;
        if (*answering_table[answering_table_idx]=='/') // a single mp3 file.
            LD3320MP3 t(answering_table[answering_table_idx]);
        else if (*answering_table[answering_table_idx]=='\\'){ // speak literally.
            MY_DBG(Serial.println(answering_table[answering_table_idx]));
            MP3Sentence t;
            t.Literal(answering_table[answering_table_idx], 2);
        }
        else {
            MY_DBG(Serial.println(answering_table[answering_table_idx]));
            MP3Sentence t(answering_table[answering_table_idx]);
        }

        switch (answering_table_idx){
            case 1: answering_table_idx=2; break;
            default: answering_table_idx=0; break;
        }
    }


    static unsigned tmoxx=micros();
    if (Timeout(12000, tmoxx) && (process_level==1)){
        if (!LD3320MP3::isOpened() && (delay(800), !LD3320MP3::isOpened())){

            LD3320::StopASR();
            nAsrStatus=LD_ASR_NONE;
            MY_DBG(Serial.println("<<<test start>>>"));

            // choose any one to test.
            static unsigned turns=0;
            if (++turns==1) test1();
            else if (turns==2) LD3320::EnterMP3ContMode(test2);
            else if (turns==3) test3();
            else if (turns<4+6) test4();

            else/* if (turns==5)*/{
                MY_DBG(Serial.println("test5 begins"));

            #if 0
            // test5, note that after run, will still go to the switch case LD_ASR_FOUNDOK.
            // so, use either the switch case or comment the switch case and test this test5.
                LD3320::LD_Reset();
                nAsrStatus=LD_ASR_BUSY;
                LD3320::LD_ChangeTable(voc_tbl0);
                LD3320::RunASR(test5_my_asr_callback);
            #else
            test5_my_asr_callback(0);
            #endif

                turns=0;
            }

        }
    }

    switch (nAsrStatus){
        case LD_ASR_IDLE:
        case LD_ASR_BUSY:
        case LD_MP3_DATA: break;
        case LD_ASR_NONE: // regularly run the asr or play an mp3 file.
            if (LD3320::RunMP3()) break; // use LD3320MP3 tmp("sample.mp3") first to trigger.
            else if (!LD3320::RunASR()){
                MY_DBG(Serial.println("ASR fault and reset and retry."));
                LD3320::LD_Reset();
                nAsrStatus=LD_ASR_NONE;
                delay(50);
            }
            break;
        case LD_ASR_FOUNDOK: // asr success.
            #if 1
            {
                unsigned b=LD3320::LD_GetResult();
                ((void(*)(unsigned))((*(voice_tables->nxtT()))[b]))(b); // call the target.
            }
            #endif
            nAsrStatus=LD_ASR_NONE;
            break;
        case LD_ASR_FOUNDZERO:
            MY_DBG(Serial.println("ASR done but no result."));
            nAsrStatus=LD_ASR_NONE;

            if (!LD3320MP3::isOpened()) answering_table_idx=5; ////// not been stopped ////////////////////////////////

            break;
        case LD_MP3_INTR: // set by mp3 intr.
            #if !(MP3_IN_ISR)
                LD3320::ProcessIntMP3();
            #endif
            break;
        case LD_MP3_REST:
            LD3320::LD_Reset();
            nAsrStatus=LD_ASR_NONE;
            break;
        default: // LD_ASR_ERROR from busy; LD_ASR_EXCEPTION or LD_ASR_INTERR from intr.
            MY_DBG(Serial.printf("Unexpected error (%u).\r\n", nAsrStatus));
            LD3320::LD_Reset();
            nAsrStatus=LD_ASR_NONE;
            delay(50);
            break;
    }

    static unsigned tmo=micros();
    if (Timeout(1350, tmo)){
        MY_DBG(Serial.printf("%u %u %u\r\n", nAsrStatus, the_debug_data, the_debug_cnt));

        // back to level 0 after about 30 seconds.
        if ((process_level==2) && ++process_level_back>16){
            LD3320::StopASR(); // it is during sampling so we have to stop it or fault.
            nAsrStatus=LD_ASR_NONE; // since stop would back to idle.
            foo_level1_common();
        }
    }

    loop_IoT();
}
