// rearranged LD3320 sample code, for tests on both asr and mp3 playing by ken woo 20220614.v.0.2.
// 20220618 v.0.3, support esp-m3 which disabled tx/rx. and with some structural modifies.
// 20220624 v.0.4, example, now uses station mode and auto connect and supports multiple files upload. add mp3 cont mode.
// 20220626 v.0.5. debug.
// 20220629 v.0.6, finalized.
// 20220629 v.0.6, example01.
//
// this example usage: 1) prepare enough filesystem space. 2) write flash by earse flash to all contents.
// 3) pc wl-connects to this ap mode and uploads mp3 files by http://192.168.x.x/upload. 4) asr-ing.
// 5) afterwards, if want to write flash, must earse only sketch otherwise the mp3 files would gone.
//
// note1: if enabled MYDB, be care of occasions of violations of playing MP3 and accessing file system via MYDB.
// note2: set ap_mode_or_sta_mode=1 to be sta, 0 to be ap.
// note3: set sta_do_the_connect=1 to demand connecting to ap.
// note4: the remote service {"motor", 8080, "tcp"} is used.
//
// 20220629 v.0.6-1, example01. fixed a bug here and added some note. add other used libs at this pack.
//
// 20220704 v.0.7, example02. ASRing.
// 20220705 v.0.8, example02. minor modified and as to re-test passed on M3(1MB fs/in folder words_simp). All files included.
// 20220709 v.0.8-2, example02. adjust-sensitivity enabled and other minor modified. note that when write flash might cause
//                                 voilation with isr and wdt-reset, disable isr while write could be alright.
//
// 20221112 v.0.8-4, example02. added an item to control voc-recog enabled or disabled, and another acronym item for reboot,
// where shows how easy to add items/commands.
// (a comment that I used 15MB FS in 16MB flash, MYDEBUG enabled, ESP8266_SPEEDUP 0,
// no InTheIram; since the SPI line noise if high speed.)




#include "LD3320.h"
#include"tables.h"

#define POLL_DEV_PRD            7000 // every 7 seconds to poll device.




//AAAAAAAAAAA/////////////////////////////////////////////////////////////////////////////////////////////////// the IoTs
#define MY_VERSION_NUMBER "[20221112.0.8-4]"

#include <ESP8266WiFi.h>
#include <ESP8266WebServer.h>
#include <ESP8266HTTPUpdateServer.h>
#include <ESP8266httpUpdate.h>
#include <ESP8266mDNS.h>
#include <ESP8266WiFiMulti.h>
#include <WiFiUdp.h>


#define MYDB 1
#if MYDB
    #include"Esp8266Vars2File.h" // if it is not used, must set ap_mode_or_sta_mode either AP(0) or STA(1).
    // for this lib being supported, the ap/sta mode is kept even poweroff.
    // and even which mode set, every 10 times of power up, the ap mode will be used once. reboot to recover.
#endif // MYDB


#define MYDEBUG 1
#if MYDEBUG
    #define MY_DBG(x) x
#else
    #define MY_DBG(x)
#endif


// -------------------------- for part of hostname --- note: this config will be a service info as well -----------------------
#define LSERVICE_1 "voice-recog" // name


// -------------------------- for default connection names ------------------------------
// i.e., http://hostname.local/
#define HOSTNAME_BASE "esp8266-iot" // for local used.
#define HOSTNAME ((String(HOSTNAME_BASE)+"-"+LSERVICE_1).c_str()) // for local used.


// -------------------------- other settings ------------------------------
// all the TMO values is restricted to no more than 35 minutes; see the Timeout() for detail. however,
// cascading could overcome.
#define TMO_IOT_LOOP_PERIOD 75 // IoT polling at each 75ms
#define TMO_STA_CONN_HEALTH_PERIOD 6000 // each 6 seconds to check the connection health by STA(while ping is alive)
#define TMO_MDNS_Q_PERIOD 30000 // each 30 seconds to query services

// [Multicast]
#define MULTICAST_ADDRESS "239.255.255.250"
#define MULTICAST_PORT "8888"
#define MULTICAST_BUFF_SIZE 15 // default is 16 in total. only (limited)for receiver.
#define MULTICAST_TR 1 // transmitting mode
#define MULTICAST_RV 2 // receiving mode


// -------------------------- main-control variables ------------------------------

const char *const WIFI_CRED[][2]={ // wifi credential
    {"ESP_TESTAP", "87654321"}, // the first one will be used which is in AP mode.
    {"Office", "123456"},
    {"Xperia X Performance_bc10", "e848acbcfd98"},
    {"etc", "etc"},
};

typedef struct smDnsRecord{ // for mdns
    const char *service;
    const unsigned port;
    const char *proto;
} smDnsRecord;

const smDnsRecord mdns_sskn[]={ // for mdns. the known services this device needs from other devices.
    {"http", 80, "tcp"},
    {"https", 443, "tcp"},
    {"test", 81, "tcp"},
    {"motor", 8080, "tcp"},
#if STA_MODE
    {"multicast", 8888, "udp"},
#endif // STA_MODE
};

bool mdns_enabled=false; // set true when connected. could be disabled once aimed and got the ip.

// for demanding STA to connect AP or not, default is yes, controlled by user.
unsigned sta_do_the_connect=0;

// for demanding this local device broadcast or not. it is supposed to switched by a http request.
unsigned multicast_is_broadcasting=0; // 0, disable; 1/MULTICAST_TR, as the transmitter; 2/MULTICAST_RV, as the receiver. and running.

// for where multicasted this device got from of the string. however, since the role can not both at the same time,
// the multicast receiver could still use this str to broadcast.
String multicast_got_or_put_str="";

// for http GET returned
String http_got_str="";

// for connected remote ap ip, in STA mode
char remote_ap_ip[16] = {0};

// remote http server ip, for doing http requests. the auto-http requests is on only if this var assigned.
char http_request_ip[16] = {0};

unsigned is_station_mode_established=0; // change bool to unsigned because elsewhere InTheIram used.

unsigned wifi_ap_credential_idx=0; // for setting up the ap credential index in WIFI_CRED. default is 0.

String ap_credential_custom_name, ap_credential_custom_pw, sta_credential_custom_name, sta_credential_custom_pw; // runtime specified.

ESP8266WebServer *server_p;

ESP8266HTTPUpdateServer httpUpdater; // webupdater, will be http://HOSTNAME.local/update.

ESP8266WiFiMulti *WiFiMulti=0; // for sta mode.

String webupd_addr=""; // the target url for this device processing remote update.

unsigned force_update=0; // force remote update to start immediately.

String uri_list=""; // for listing the uris this device provided.

String upload_abort_filename=""; // for storing the file name which broken when upload.

String my_local_ip=""; // for storing the station mode local ip we got.

unsigned ap_mode_or_sta_mode=0; // be ap or sta, default is ap. could be changed the role by http request /tosta or toap.


// -------------------------- ap mode ------------------------------

bool AP_enter_ap_mode(){ // use WIFI_CRED. default is wifi_ap_credential_idx=0. idx can be changed by http request of /getap and /setap.
    if (ap_credential_custom_name!=""){
        WiFi.softAP(ap_credential_custom_name.c_str(), ap_credential_custom_pw.c_str());
        IPAddress myIP = WiFi.softAPIP();
        MY_DBG(Serial.print("AP IP address: "));
        MY_DBG(Serial.println(myIP));
        return true;
    }
    else if (wifi_ap_credential_idx<sizeof(WIFI_CRED)/sizeof(WIFI_CRED[0])){
        WiFi.softAP(WIFI_CRED[wifi_ap_credential_idx][0], WIFI_CRED[wifi_ap_credential_idx][1]);
        IPAddress myIP = WiFi.softAPIP();
        MY_DBG(Serial.print("AP IP address: "));
        MY_DBG(Serial.println(myIP));
        return true;
    }
    return false;
}


// -------------------------- station mode ------------------------------

bool STA_enter_station_mode(){ // 12 seconds waiting
    unsigned char u=0;

    // Connect to WiFi network
    MY_DBG(Serial.println(""));
    MY_DBG(Serial.println("[-= Enter Station Mode =-]"));
    MY_DBG(Serial.print("WiFi Connecting..."));

    // correct order
    WiFi.disconnect(); // since cases could WiFi reconnect.
    WiFi.mode(WIFI_OFF); // come up with disconnect, not sure ok or not.
    WiFi.mode(WIFI_STA);
    WiFi.begin();
    WiFi.hostname(HOSTNAME); // set after WiFi.begin().

    if (!WiFiMulti){
        WiFiMulti=new ESP8266WiFiMulti;
        if (!WiFiMulti) return false;

        if (sta_credential_custom_name!="")
            WiFiMulti->addAP(sta_credential_custom_name.c_str(), sta_credential_custom_pw.c_str());

        for (int i=0; i<(sizeof(WIFI_CRED)/sizeof(WIFI_CRED[0])); i++)
            WiFiMulti->addAP(WIFI_CRED[i][0], WIFI_CRED[i][1]);
    }

    while (WiFiMulti->run() != WL_CONNECTED) {
        delay(200);
        MY_DBG(Serial.print("."));
        if (u++>60){ // 12 seconds, might not enough
            MY_DBG(Serial.println(" WiFi connection failed..."));
            return false;
        }
    }
    MY_DBG(Serial.println("WiFi connected"));

    // Print the remote MAC address
    uint8_t macAddr[6];
    WiFi.macAddress(macAddr);
    MY_DBG(Serial.printf("Remote MAC adr: %02X:%02X:%02X:%02X:%02X:%02X\r\n",\
        macAddr[0], macAddr[1], macAddr[2], macAddr[3], macAddr[4], macAddr[5]));

    ip_info info;
    wifi_get_ip_info(STATION_IF, &info);
    MY_DBG(Serial.printf("IP ADDR: " IPSTR "\r\n", IP2STR(&info.ip)));
    MY_DBG(Serial.printf("GATEWAY: " IPSTR "\r\n", IP2STR(&info.gw)));
    MY_DBG(Serial.printf("NETMASK: " IPSTR "\r\n", IP2STR(&info.netmask)));

    // resolve the remote ip
    uint32 x=*(uint32*)(&info.ip)&*(uint32*)(&info.netmask);
    ip_addr y=*(ip_addr*)(&x);
    snprintf(remote_ap_ip, 16, "%d.%d.%d.%d",\
        ip4_addr1(&y)? ip4_addr1(&y): 1,\
        ip4_addr2(&y)? ip4_addr2(&y): 1,\
        ip4_addr3(&y)? ip4_addr3(&y): 1,\
        ip4_addr4(&y)? ip4_addr4(&y): 1
        );
    MY_DBG(Serial.print(F("Remote AP IP: ")));
    MY_DBG(Serial.println(remote_ap_ip));

    MY_DBG(Serial.printf("Default local hostname: %s\r\n", WiFi.hostname().c_str()));
    MY_DBG(Serial.print("DNS #1, #2 IP: "));
    MY_DBG(WiFi.dnsIP().printTo(Serial));
    MY_DBG(Serial.print(", "));
    MY_DBG(WiFi.dnsIP(1).printTo(Serial));
    MY_DBG(Serial.println());
    MY_DBG(Serial.printf("RSSI: %d dBm\r\n", WiFi.RSSI()));
    MY_DBG(WiFi.printDiag(Serial));
    
    // Print the local IP address
    MY_DBG(Serial.println(""));
    MY_DBG(Serial.print("IP is assigned: "));
    MY_DBG(Serial.println(WiFi.localIP()));
    my_local_ip=WiFi.localIP().toString();
    return true;
}

void STA_drop_station_mode(){
    WiFi.disconnect();
    WiFi.mode(WIFI_OFF); // no this line will wdt-reset
}


// -------------------------- http server ------------------------------

ESP8266WebServer* HTTPH_create_http_server(){ // ESP8266WebServer & MDNS & Updater facilities
    // Create an instance of the server
    // specify the 80 port to listen on as an argument
    ESP8266WebServer *a=new ESP8266WebServer(80);

    // setup the mdns.
    if (!MDNS.begin(HOSTNAME)) { // Start the mDNS responder
        MY_DBG(Serial.println("Error setting up MDNS responder!"));
    }
    MY_DBG(Serial.println("mDNS responder started"));

    // setup the updater; which is http://hostname/update.
    httpUpdater.setup(a);

    // Start the http server
    a->begin();

    // config mdns.
    MDNS.addService(LSERVICE_1, "tcp", 80);

    MY_DBG(Serial.println(String("UPDATE page is http://")+HOSTNAME+".local:80/update"));
    MY_DBG(Serial.println("HTTP Server started"));
    return a;
}

void remove_the_http_server(){
    if (server_p) delete server_p;
    server_p=0;
}


// -------------------------- http client ------------------------------

bool HTTPC_http_client_connect(const char *url){
    WiFiClient client;
    HTTPClient http; // must be declared after WiFiClient for correct destruction order, because used by http.begin(client,...)
    bool returned_state=false; // true only if got data.

    // configure server and url
    http.begin(client, url);

    // start connection and send HTTP header
    int httpCode = http.GET();
    if (httpCode > 0) {
        // HTTP header has been send and Server response header has been handled

        // file found at server
        if (httpCode == HTTP_CODE_OK) {

            // get lenght of document (is -1 when Server sends no Content-Length header)
            int len = http.getSize();

            #if 0 // with API

                http_got_str=http.getString();
                MY_DBG(Serial.println(http_got_str));

            #else // or "by hand"

                // create buffer for read
                uint8_t buff[128] = {0};
                http_got_str="";

                // get tcp stream
                WiFiClient * stream = &client;

                // read all data from server
                while (http.connected() && (len > 0 || len == -1)) {
                    // read up to 128 byte
                    int c = stream->readBytes(buff, std::min((size_t)len, sizeof(buff)-1));
                    if (!c) {
                        MY_DBG(Serial.println("[read timeout]"));
                        http_got_str="";
                        returned_state=false;
                        break; // must have or it will stuck here! and then if stuck at GET(), set timeout
                    }

                    buff[c]=0;
                    http_got_str+=(char*)buff;

                    if (len > 0) len -= c;
                    returned_state=true;
                }

            #endif // or "by hand"

        }
        else http_got_str="";
    }
    else {
        http_got_str="";
        MY_DBG(Serial.printf("[HTTP GET... failed, error: %s]\r\n", http.errorToString(httpCode).c_str()));
    }
    http.end();
    return returned_state;
}


// -------------------------- mdns ------------------------------

unsigned mdns_sskn_len=sizeof(mdns_sskn)/sizeof(smDnsRecord);

// the count is to pause the auto query for explicitly manual query.
unsigned mDnsQuery_count=0; // the manual query result on how many same service devices were found.

void mDnsQuery(){ // unable to query the provided itselves. especially note that which can be queried should use (USE_MDNS_FULL)
    if (mDnsQuery_count) return;
    MY_DBG(printf("Sending mDNS query\r\n"));
    for (int i=0; i<mdns_sskn_len; i++){
        int n = MDNS.queryService(mdns_sskn[i].service, mdns_sskn[i].proto); // Send out query for those tcp services
        if (!n) MY_DBG(printf("no service [%s] found.\r\n", mdns_sskn[i].service));
        else {
            MY_DBG(printf("service(s) found\r\n"));
            for (int j = 0; j < n; j++)
                MY_DBG(printf("[%s] %d: %s (%s: %d)\r\n",\
                    mdns_sskn[i].service, j+1, MDNS.hostname(j).c_str(), MDNS.IP(j).toString().c_str(), MDNS.port(j)));
        }
    }
    MY_DBG(printf("mDNS query done\r\n"));
}

unsigned mDnsQueryStart(const char *serv_name, const char *serv_protocol){ // start/cont/end are bundled use together.
    // return the number of devices found.
    return mDnsQuery_count=MDNS.queryService(serv_name, serv_protocol);
}

bool mDnsQueryCont(unsigned idx, IPAddress &ip, unsigned &port, String &hostname){ // start/cont/end are bundled use together.
    // use idx to get the found. idx is from 0 to mDnsQuery_count-1.
    if (idx>=mDnsQuery_count) return false;
    hostname=MDNS.hostname(idx);
    ip=MDNS.IP(idx);
    port=MDNS.port(idx);
    return true;
}

void mDnsQueryEnd(){ // start/cont/end are bundled use together.
    mDnsQuery_count=0;
}


// -------------------------- multicast wrapped; multicast is only for sta mode ------------------------------

WiFiUDP Udp;
IPAddress addressMulticast=0;
IPAddress ifaddrMulticast=0;
unsigned portMulticast=0;
char multicast_buffer[MULTICAST_BUFF_SIZE+1]; // default is 16 in total. only (limited)for receiver.

void beginMulticastTransmitter(const char *mcast_address=MULTICAST_ADDRESS, const char *port=MULTICAST_PORT){
    // the timing and contents and tr or recv all are controlled by client code.
    if (ap_mode_or_sta_mode){ // sta mode.
        ifaddrMulticast=WiFi.localIP();
        addressMulticast=ipaddr_addr(mcast_address);
        portMulticast=String(port).toInt();
    }
}

void beginMulticastReceiver(const char *mcast_address=MULTICAST_ADDRESS, const char *port=MULTICAST_PORT){
    // the timing and contents and tr or recv all are controlled by client code.
    if (ap_mode_or_sta_mode){ // sta mode.
        ifaddrMulticast=WiFi.localIP();
        addressMulticast=ipaddr_addr(mcast_address);
        portMulticast=String(port).toInt();
        Udp.beginMulticast(ifaddrMulticast, addressMulticast, portMulticast);
    }
}

void transmitMulticast(){ // the timing and contents and tr or recv all are controlled by client code.
    if (ap_mode_or_sta_mode && (multicast_is_broadcasting==MULTICAST_TR)){
        Udp.beginPacketMulticast(addressMulticast, portMulticast, ifaddrMulticast);
        Udp.write(multicast_got_or_put_str.c_str());
        Udp.endPacket();
    }
}

bool receiveMulticast(){ // the timing and contents and tr or recv all are controlled by client code.
    if (ap_mode_or_sta_mode && (multicast_is_broadcasting==MULTICAST_RV)){
        int packetLength=Udp.parsePacket();
        if (packetLength>0){
            if (packetLength>MULTICAST_BUFF_SIZE) packetLength=MULTICAST_BUFF_SIZE;
            Udp.read(multicast_buffer, packetLength);
            multicast_buffer[packetLength]=0;
            multicast_got_or_put_str=multicast_buffer;
            ///// use IPAddress destinationIP() to identify it is for multicast address.
            ///// use IPAddress remoteIP() to identify the remote address.
            ///// use uint16_t remotePort() to identify the remote port.
            return true;
        }
    }
    return false;
}

void stopMulticast(){ // if stop, multicast_is_broadcasting must reassign if reuse.
    // the timing and contents and tr or recv all are controlled by client code.

    if (ap_mode_or_sta_mode){
        multicast_is_broadcasting=0;
        Udp.stop(); // especially note that the receiver is of concerns, so better to stop if not used.
    }
}


// -------------------------- Remote updater ------------------------------
// the device can not update itself because this updater has yet not found any exit point.

void gbl_update_started() {
  MY_DBG(Serial.println("CALLBACK:  HTTP update process started"));
}

void gbl_update_finished() {
  MY_DBG(Serial.println("CALLBACK:  HTTP update process finished"));
}

void gbl_update_progress(int cur, int total) {
  MY_DBG(Serial.printf("CALLBACK:  HTTP update process at %d of %d bytes...\n", cur, total));
}

void gbl_update_error(int err) {
  MY_DBG(Serial.printf("CALLBACK:  HTTP update fatal error code %d\n", err));
}

void gbl_updater(const char *url) {
    // wait for WiFi connection

    WiFiClient client;

    // The line below is optional. It can be used to blink the LED on the board during flashing
    // The LED will be on during download of one buffer of data from the network. The LED will
    // be off during writing that buffer to flash
    // On a good connection the LED should flash regularly. On a bad connection the LED will be
    // on much longer than it will be off. Other pins than LED_BUILTIN may be used. The second
    // value is used to put the LED on. If the LED is on with HIGH, that value should be passed
    // ESPhttpUpdate.setLedPin(LED_BUILTIN, LOW);

    // Add optional callback notifiers
    ESPhttpUpdate.onStart(gbl_update_started);
    ESPhttpUpdate.onEnd(gbl_update_finished);
    ESPhttpUpdate.onProgress(gbl_update_progress);
    ESPhttpUpdate.onError(gbl_update_error);

    t_httpUpdate_return ret = ESPhttpUpdate.update(client, url);
    // Or:
    // t_httpUpdate_return ret = ESPhttpUpdate.update(client, "server", 80, "file.bin");

    switch (ret) {
        case HTTP_UPDATE_FAILED:
            MY_DBG(Serial.printf("HTTP_UPDATE_FAILD Error (%d): %s\r\n", ESPhttpUpdate.getLastError(), ESPhttpUpdate.getLastErrorString().c_str()));
            break;
        case HTTP_UPDATE_NO_UPDATES:
            MY_DBG(Serial.println("HTTP_UPDATE_NO_UPDATES"));
            break;
        case HTTP_UPDATE_OK:
            MY_DBG(Serial.println("HTTP_UPDATE_OK"));
            break;
    }
}


// -------------------------- for files upload, uri: /upload. and the list dir/del file facilities. ---------------------------
String get_content_type(String filename){
    if (filename.endsWith(".htm")) return "text/html";
    else if (filename.endsWith(".html")) return "text/html";
    else if (filename.endsWith(".css")) return "text/css";
    else if (filename.endsWith(".js")) return "application/javascript";
    else if (filename.endsWith(".png")) return "image/png";
    else if (filename.endsWith(".gif")) return "image/gif";
    else if (filename.endsWith(".jpg")) return "image/jpeg";
    else if (filename.endsWith(".ico")) return "image/x-icon";
    else if (filename.endsWith(".xml")) return "text/xml";
    else if (filename.endsWith(".pdf")) return "application/x-pdf";
    else if (filename.endsWith(".zip")) return "application/x-zip";
    else if (filename.endsWith(".gz")) return "application/x-gzip";
    else if (filename.endsWith(".bin")) return "application/octet-stream";
    else if (filename.endsWith(".mp3")) return "application/octet-stream";
    return "text/plain";
}

bool page_handleFileRead(String path){ // send the file to the client if it exists
    LittleFS.begin();
    if (path.endsWith("/")) path += "index.html"; // If a folder is requested, send the index file
    String pathWithGz = path + ".gz";
    if (LittleFS.exists(pathWithGz) || LittleFS.exists(path)) { // If the file exists, either as a compressed archive, or normal
        if (LittleFS.exists(pathWithGz)) path = pathWithGz; // If there's a compressed version available, then use the compressed version
        File file = LittleFS.open(path, "r"); // Open the file
        size_t sent = server_p->streamFile(file, get_content_type(path)); // Send it to the client
        file.close(); // Close the file
        MY_DBG(Serial.println(String("Sent the file done: ") + path));
        LittleFS.end();
        return true;
    }
    MY_DBG(Serial.println(String("File Not Found: ") + path));
    // If the file doesn't exist, return false and direct back to notify false page.

    LittleFS.end();
    return false;
}

void page_handleFileUpload_cb(){

    extern String upload_abort_filename;
    upload_abort_filename="";

    static File fsu_file;
    HTTPUpload &upload = server_p->upload();
    String filename = upload.filename;
    switch (upload.status){
        case UPLOAD_FILE_START:
            LittleFS.begin();
            if (!filename.startsWith("/")) filename = "/"+filename;
            MY_DBG(printf("handle the file upload: %s\r\n", filename.c_str()));
            fsu_file = LittleFS.open(filename, "w"); // Open the file for writing in littlefs (create if it doesn't exist)
            break;
        case UPLOAD_FILE_WRITE:
            if (fsu_file)
                fsu_file.write(upload.buf, upload.currentSize); // Write the received bytes to the file
            break;
        case UPLOAD_FILE_END:
            if (fsu_file) { // If the file was successfully created
                fsu_file.close(); // Close the file
                MY_DBG(printf("done the file, total size: %u\r\n", upload.totalSize));
                server_p->sendHeader("Location", "/success.html"); // Redirect the client to the success page
                server_p->send(303);
            }
            else {
                server_p->send(500, "text/plain", "500: could not create file");
                MY_DBG(printf("error, could not create file: %s\r\n", filename.c_str()));
            }
            LittleFS.end();
            break;
        default:
            LittleFS.end();
            MY_DBG(printf("upload error. code: %d, fn: %s\r\n", upload.status, filename.c_str()));
            upload_abort_filename=filename;
            break;
    }
}

void Ls(String *s=0, const char *path="/"){
    LittleFS.begin();

    Dir dir = LittleFS.openDir(path);
    int i=0, j=0;

    if (s){
        (*s)+="list directory \"";
        (*s)+=path;
        (*s)+="\":\r\n--------\r\n";
    }
    else MY_DBG(printf("list directory \"%s\":\r\n--------\r\n", path));
    while (dir.next()){
        i++;
        if (s){
            (*s)+=dir.fileName();
            (*s)+="    ";
            (*s)+=(dir.isDirectory()? (j++, "<DIR>"): (String(dir.fileSize())+" bytes").c_str());
            (*s)+="\r\n";
        }
        else
            MY_DBG(printf("%s    %s\r\n",\
                dir.fileName().c_str(), dir.isDirectory()? (j++, "<DIR>"): (String(dir.fileSize())+" bytes").c_str()));
    }
    if (s){
        (*s)+="    ";
        (*s)+=j;
        (*s)+="  Dirs    ";
        (*s)+=(i-j);
        (*s)+="  Files\r\n";
    }
    else MY_DBG(printf("    %d  Dirs    %d  Files\r\n", j, i-j));

    FSInfo fs_info;
    if (LittleFS.info(fs_info)){
        if (s){
            (*s)+="\r\nentire space ";
            (*s)+=fs_info.totalBytes;
            (*s)+=" bytes\r\n";
            (*s)+=fs_info.usedBytes;
            (*s)+=" bytes used\r\nfree space ";
            (*s)+=fs_info.totalBytes-fs_info.usedBytes;
            (*s)+=" bytes\r\n--------\r\n";
        }
        else MY_DBG(printf("\r\nentire space %d bytes\r\n%d bytes used\r\nfree space %d bytes\r\n--------\r\n",\
            fs_info.totalBytes, fs_info.usedBytes, fs_info.totalBytes-fs_info.usedBytes));
    }

    LittleFS.end();
}


// -------------------------- other functions ------------------------------
bool Timeout(unsigned ms, unsigned &store_start_time_us, bool one_shot=false, bool one_shot_lasting=false){
    if (one_shot && !store_start_time_us) return one_shot_lasting;
    if (int(micros())-int(ms*1000+store_start_time_us)>=0){ ///// this is ok
        store_start_time_us=one_shot? 0: micros();
        return true;
    }
    return false;
}


// -------------------------- main controls ------------------------------

void the_http_request(){

    extern void my_http_request(); // all the http requests from this device could be controlled in this function.
    if (*http_request_ip) my_http_request(); // called only if http_request_ip assigned.

}

void the_http_server(){

    server_p=HTTPH_create_http_server();

    server_p->on("/", [](){
        server_p->send(200, "text/html", ap_mode_or_sta_mode? "<h1>WiFi STATION</h1>": "<h1>WiFi AP</h1>");
    });

    server_p->on("/version", []() {server_p->send(200, "text/plain", MY_VERSION_NUMBER);});

    server_p->on("/reset", []() {
        server_p->send(200, "text/plain", "now reset...");
        delay(1000);
        while (1);
    });

    server_p->on("/s", []() {server_p->send(200, "text/plain", String(random(10)+1).c_str());});

    server_p->on("/getap", [](){
        String a("choose the number:\r\n");
        for (int i=0; i<(sizeof(WIFI_CRED)/sizeof(WIFI_CRED[0])); i++){
            a+=String(i+1)+". ["+WIFI_CRED[i][0]+", "+/*WIFI_CRED[i][1]*/"***"+"]\r\n";
        }
        a=a+"currently is "+(wifi_ap_credential_idx+1)+". ["+WIFI_CRED[wifi_ap_credential_idx][0]+"]\r\n";
        server_p->send(200, "text/plain", a.c_str());
    });

    server_p->on("/setap", [](){ // ?n=<i>.
        String n=server_p->arg("n");
        int nn=n.toInt()-1;
        if (nn>=0 && nn<sizeof(WIFI_CRED)/sizeof(WIFI_CRED[0])) wifi_ap_credential_idx=nn;
        else wifi_ap_credential_idx=0;

        #if MYDB
            cDataBase("/settings").ReplaceWordsEx("wifi_ap_credential_idx", 0, wifi_ap_credential_idx);
        #endif // MYDB

        server_p->send(200, "text/plain", (String(WIFI_CRED[wifi_ap_credential_idx][0])+" chosen.").c_str());
    });

    server_p->on("/toap", [](){
        server_p->send(200, "text/plain", "done.");
        ap_mode_or_sta_mode=0;

        #if MYDB
            cDataBase("/settings").ReplaceWordsEx("ap_mode_or_sta_mode", 0, ap_mode_or_sta_mode);
        #endif // MYDB

        delay(1000);
        while (1); // reset to change role.
    });

    server_p->on("/tosta", [](){
        server_p->send(200, "text/plain", "done.");
        ap_mode_or_sta_mode=1;

        #if MYDB
            cDataBase("/settings").ReplaceWordsEx("ap_mode_or_sta_mode", 0, ap_mode_or_sta_mode);
        #endif // MYDB

        delay(1000);
        while (1); // reset to change role.
    });

    server_p->on("/credential", [](){ // ?r=<the role is ap or sta>&n=<name>&p=<password>.
        server_p->send(200, "text/plain", "done.");
        String role=server_p->arg("r");
        String cname=server_p->arg("n");
        String pw=server_p->arg("p");

        #if MYDB
            unsigned n_len=cname.length()+1, p_len=pw.length()+1;
            if (role=="ap"){
                cDataBase("/settings").ReplaceBytesEx("ap_credential_custom_name", 0, cname.c_str(), n_len, n_len);
                cDataBase("/settings").ReplaceBytesEx("ap_credential_custom_pw", 0, pw.c_str(), p_len, p_len);
            }
            else if (role=="sta"){
                cDataBase("/settings").ReplaceBytesEx("sta_credential_custom_name", 0, cname.c_str(), n_len, n_len);
                cDataBase("/settings").ReplaceBytesEx("sta_credential_custom_pw", 0, pw.c_str(), p_len, p_len);
            }
        #endif // MYDB

    });

    // assign the update website url to this device(it is intended for this local device connecting to)
    server_p->on("/webupdinfo", []() { // http://<ip address>/webupdinfo?addr=<remote url; e.g., http://192.168.1.1/a.bin>.
        String addr=server_p->arg("addr");
        if (addr=="") server_p->send(404, "text/plain", "0. page not found.");
        else {
            webupd_addr=addr;
            server_p->send(200, "text/plain", ("1. accepted: "+webupd_addr).c_str());
        }
    });

    // force global update immediately
    server_p->on("/forceupdate", []() {
        force_update=1;
        server_p->send(200, "text/plain", String("1. accepted. on 10 seconds to start updating...").c_str());
    });

    server_p->on("/listcmds", []() { // list all commands.
        server_p->send(200, "text/plain", uri_list.c_str());
    });

    // list directories.
    server_p->on("/ls", []() {
        String a;
        Ls(&a);
        server_p->send(200, "text/plain", a.c_str());
    });

    // delete a file.
    server_p->on("/del", []() { // ?n=<filename>.
        String n=server_p->arg("n");
        LittleFS.begin();
        LittleFS.remove(n);
        LittleFS.end();
        server_p->send(200, "text/plain", "done");
    });

    server_p->on("/success.html", HTTP_GET, []() {

        extern String upload_abort_filename;

        server_p->send(
            200, "text/html",\
            (upload_abort_filename=="")\
                ? "<!DOCTYPE HTML>\r\n<html>\r\nsuccess\r\n</html>"\
                : (String("<!DOCTYPE HTML>\r\n<html>\r\nabort at file: ")+\
                    upload_abort_filename+\
                    String("\r\n</html>")).c_str()\
        );
    });

    // upload a file to this device. POST.
    server_p->on("/upload", HTTP_POST, []() {
            server_p->send(200);
        },
        page_handleFileUpload_cb
    );

    // upload a file to this device. GET.
    server_p->on("/upload", HTTP_GET, []() {
        if (!page_handleFileRead("/upload.html")){
            server_p->send(200, "text/html", "<!DOCTYPE HTML>\r\n<html>\r\n\
                <form method=\"post\" enctype=\"multipart/form-data\">\
                <input type=\"file\" name=\"name\" multiple>\
                <input class=\"button\" type=\"submit\" value=\"Upload\"></form>\r\n\
                </html>");
        }
    });

    // use uri_list+=str; to expand uri list.
    uri_list=\
        "/listcmds\r\n"\
        "/\r\n"\
        "/version\r\n"\
        "/reset\r\n"\
        "/s\r\n"\
        "/update (by selecting an update file from pc and upload to update)\r\n"\
        "/webupdinfo?addr=<remote url; e.g., http://192.168.1.1/a.bin>\r\n"\
        "/forceupdate\r\n"\
        "/getap\r\n"\
        "/setap?n=<i>\r\n"\
        "/credential?r=<the role is ap or sta>&n=<name>&p=<password>\r\n"\
        "/toap\r\n"\
        "/tosta\r\n"\
        "/ls (list root files and directories)\r\n"\
        "/del?n=<filename; delete this file>\r\n"\
        "/upload\r\n";
}

void setup_IoT(){
    if (ap_mode_or_sta_mode) MY_DBG(Serial.println("station mode entered."));
    else {
        MY_DBG(Serial.println("Configuring access point..."));
        AP_enter_ap_mode();
        mdns_enabled=true;

        the_http_server(); // setup the http server
    }
}


// AAAA ------------------------------------------new added for my application, of the IoTs.
bool poll_device=true;
const char *device_name="motor"; // protocol tcp, port 8080.
String target_dev_ip="", target_dev_last_sent="";

unsigned successive_cmd_issue=0; // for http.
int successive_cmd_status=0; // for code flow.

void loop_Remote(){
    if (poll_device){ // device still not found and needs to poll.
        static unsigned wtmp=0;
        static unsigned tmoz=micros();
        if (!Timeout(POLL_DEV_PRD, tmoz)) return; // to poll device

        IPAddress ip;
        unsigned port;
        String a;
        int i=mDnsQueryStart(device_name, "tcp");
        for (int j=0; j<i; j++){
            if (mDnsQueryCont(j, ip, port, a)){
                target_dev_ip=ip.toString();

                #if MYDB
                    unsigned n_len=target_dev_ip.length()+1;
                    cDataBase("/settings").ReplaceBytesEx("prior_target_dev_ip", 0, target_dev_ip.c_str(), n_len, n_len);
                #endif // MYDB

                poll_device=false;
                break;
            }
        }

        // try 10 times and still not found device, use former one. since the remote might disable mDns in a while.
        if (poll_device && wtmp++>10){
            wtmp=0;
            if (target_dev_ip!="") poll_device=false; // the string is loaded while bootup.
        }

    }
    else {
        static unsigned xtmp=1;
        switch (successive_cmd_issue){ // for the -1 to be handled elsewhere.
            case 1: // syncoff
                HTTPC_http_client_connect(("http://"+target_dev_ip+"/params/mtr/syncoff").c_str());
                target_dev_last_sent=http_got_str;
            case 2: // open
                HTTPC_http_client_connect(("http://"+target_dev_ip+"/params/gate/open").c_str());
                target_dev_last_sent=http_got_str;
                break;
            case 3: // close
                HTTPC_http_client_connect(("http://"+target_dev_ip+"/params/gate/close").c_str());
                target_dev_last_sent=http_got_str;
                break;
            case 4: // speed
                HTTPC_http_client_connect(("http://"+target_dev_ip+"/s").c_str());
                target_dev_last_sent=http_got_str;
                break;
            case 5: // probe
                HTTPC_http_client_connect(("http://"+target_dev_ip+"/params/mtr/probe").c_str());
                target_dev_last_sent=http_got_str;
                break;
            case 6: // syncon
                HTTPC_http_client_connect(("http://"+target_dev_ip+"/params/mtr/syncon").c_str());
                delay(800);
                HTTPC_http_client_connect(("http://"+target_dev_ip+"/params/mtr/on").c_str());
                target_dev_last_sent=http_got_str;
                break;
            case 7: // small-open
                HTTPC_http_client_connect(("http://"+target_dev_ip+"/params/gate/efficiency").c_str());
                target_dev_last_sent=http_got_str;
                break;
            case 8: // large-open
                HTTPC_http_client_connect(("http://"+target_dev_ip+"/params/gate/performance").c_str());
                target_dev_last_sent=http_got_str;
                break;
            case 9: // inc-open
                if (xtmp){
                    xtmp=0;
                    HTTPC_http_client_connect(("http://"+target_dev_ip+"/params/mtr/syncoff").c_str());
                    delay(800);
                }
                HTTPC_http_client_connect(("http://"+target_dev_ip+"/t?t=3").c_str());
                target_dev_last_sent=http_got_str;
                break;
            case 10: // inc-close
                if (xtmp){
                    xtmp=0;
                    HTTPC_http_client_connect(("http://"+target_dev_ip+"/params/mtr/syncoff").c_str());
                    delay(800);
                }
                HTTPC_http_client_connect(("http://"+target_dev_ip+"/t?t=4").c_str());
                target_dev_last_sent=http_got_str;
                break;
            case 11: // loose-close
                HTTPC_http_client_connect(("http://"+target_dev_ip+"/params/gate/looseclose").c_str());
                target_dev_last_sent=http_got_str;
                break;
            default: if (successive_cmd_issue==-1) return; break; // for the -1 to be handled elsewhere.
        }
        successive_cmd_issue=0;
    }
}
// VVVV ------------------------------------------new added for my application, of the IoTs.


void loop_IoT(){
    static unsigned tmo1=micros();
    if (!Timeout(TMO_IOT_LOOP_PERIOD, tmo1)) return; // polling at each 75ms

    if (ap_mode_or_sta_mode){ // station mode.
        if (!sta_do_the_connect){
            if (is_station_mode_established){
                is_station_mode_established=0;
                STA_drop_station_mode();
            }
            return; // here will return if no station connection, so latter code would not run.
        }
        static unsigned tmo2=micros(), tmo02=tmo2; // tmo02, after 3 seconds starts to connect.
        if (Timeout(3000, tmo02, true) || Timeout(TMO_STA_CONN_HEALTH_PERIOD, tmo2, true, true)){
            if (!is_station_mode_established){
                mdns_enabled=is_station_mode_established=STA_enter_station_mode(); // it is blocking wait.
                static unsigned fault_retry_cnt=0;
                if (is_station_mode_established){ // run at here since mdns needs the ip or wdt-reset.
                    fault_retry_cnt=0;

                    remove_the_http_server();
                    the_http_server(); // setup the http server

                }
                else { // a fault case that resulting wdt reset and then sta always invalid, so wdt reset again to retry!!!!!
                    tmo2=micros();
                    if (fault_retry_cnt++>3){fault_retry_cnt=0; sta_do_the_connect=0;} //// need to abort or it block everything.
                    ////if (fault_retry_cnt++>25) while (1);
                }
                return;
            }
        }
        else return;


// AAAA -------------------------------------new added for my application, of the IoTs.
        switch (successive_cmd_status){ // 0, idle; set to 1 to start to code flow. set to 10 to reset and back to 0.
            case 1: // when demand ap-connection, set status to 1 to polling wifi-conn-status.
                if (is_station_mode_established) successive_cmd_status=2; // on to line.
                else if (!sta_do_the_connect) successive_cmd_status=10;
                break;
            case 3: // searching device phase.
                loop_Remote();
                if (!poll_device) successive_cmd_status=4; // the remote device connected.
                break;
            case 5: // accessing device phase.
                loop_Remote();
                break;
            case 10: // reset all.
                poll_device=1;
                sta_do_the_connect=0;
                successive_cmd_status=0;
                break;
            default: break;
        }
// VVVV ---------------------------------new added for my application, of the IoTs.


    }

    if (mdns_enabled){
        MDNS.update();

        //// static unsigned tmo4=micros();
        //// if (Timeout(TMO_MDNS_Q_PERIOD, tmo4)) mDnsQuery(); // each 30 seconds to query services.

    }

    server_p->handleClient(); // handles http requests from remotes.

    if (force_update && webupd_addr!=""){ // handles update.
        static unsigned tmo3=micros();
        if (force_update==1){
            force_update=2;
            tmo3=micros();
        }
        else if ((force_update==2) && Timeout(10000, tmo3, true)){ // 10 seconds to start
            force_update=0;
            gbl_updater(webupd_addr.c_str());
        }
    }

    the_http_request(); // for local http request to remote
}

void my_http_request(){;}


// -------------------------- main loop ------------------------------

void the_iot_setup(){
    #if MYDB
        unsigned boot_up_times=0;
        unsigned char buff[32];
        unsigned tmp=0;
        cDataBase db("/settings");
        db.Open();
        db.ReadWords(boot_up_times, "boot_up_times");
        db.ReadWords(wifi_ap_credential_idx, "wifi_ap_credential_idx");
        db.ReadWords(ap_mode_or_sta_mode, "ap_mode_or_sta_mode");
        db.ReadWords(sta_do_the_connect, "sta_do_the_connect");
        if (db.ReadWords(tmp, "set_volume")) set_volume=tmp;
        if (db.ReadWords(tmp, "asr_sensitive_l")) asr_sensitive_l=tmp;

        buff[0]=0; db.ReadData(buff, 32, "ap_credential_custom_name", 0); ap_credential_custom_name=(char*)buff;
        buff[0]=0; db.ReadData(buff, 32, "ap_credential_custom_pw", 0); ap_credential_custom_pw=(char*)buff;
        buff[0]=0; db.ReadData(buff, 32, "sta_credential_custom_name", 0); sta_credential_custom_name=(char*)buff;
        buff[0]=0; db.ReadData(buff, 32, "sta_credential_custom_pw", 0); sta_credential_custom_pw=(char*)buff;
        buff[0]=0; db.ReadData(buff, 32, "prior_target_dev_ip", 0); target_dev_ip=(char*)buff;

        db.Close();

        ++boot_up_times;
        cDataBase("/settings").ReplaceWordsEx("boot_up_times", 0, boot_up_times);
        if (!(boot_up_times%10)){
            cDataBase("/settings").Renew();

            ap_mode_or_sta_mode=0; // force to in ap mode, reboot to recover.
        }
    #endif // MYDB

    setup_IoT();
}
//VVVVVVVVVVV/////////////////////////////////////////////////////////////////////////////////////////////////// the IoTs
/*
連線狀態 lian xian zhuang tai
取消 qu xiao
停止 ting zhi
終 zhong
斷 duan
下線 xia xian
連 lian
確定 que ding
重置 chong zhi
弄 long
頁思 ye si
否 pi
素 su
放棄 fang qi
退出 tui chu
忽略 hu lue
重新開機 chong xin kai ji
啟動 qi dong
播歌 bo ge
放歌 fang ge
史爹遜 shi die xun
惡劣 e lie
晃戲 huang hu
系係 ji xi
侯 hou
好的 hao di
設定儲存 she ding chu cun
成為自動 cheng wei zi dong
半全睡眠 ban quan shui mian
開關機網路 kai guan ji wang lu
暫時間 zan shi jian
語音識別 yu yin shi bie
調整 tiao zheng
恢復增加減少零敏度 hui fu zeng jia jian shao ling min duo
*/
//AAAAAAAAAAA/////////////////////////////////////////////////////////////////////////////////////////////////// LD3320




// ---------------------------------------------------------------------------------------- wdt/IoT halt.
IRAM_ATTR volatile unsigned &HardwareWDT = *((volatile unsigned*)0x60000900);
IRAM_ATTR volatile unsigned &HwTimer1Load = *((volatile unsigned*)0x60000600);
#define disableHardwareWDT  (HardwareWDT&=~1)
#define enableHardwareWDT   (HardwareWDT|=1)

void halt_wdt(){
    static unsigned t1=1;
    if (t1){
        t1=0;
        WiFi.mode(WIFI_OFF);
        WiFi.forceSleepBegin(-1);
        system_soft_wdt_feed();
        system_soft_wdt_stop();
        disableHardwareWDT;
        delay(1500);
        MY_DBG(Serial.printf("\r\nHALT IOT and WDT\r\n"));
    }
}


// ---------------------------------------------------------------------------------------- issuing remote commands aux functions.
unsigned RemoteAccsAux(unsigned x=0){ // only if either x nonzero or has issue will modify mp3_response_local_custom non-null.
    extern mytable answering_table;
    extern String mp3_response_local_custom;
    String a="";
    if (poll_device && !x) a=answering_table[47]; // " oops"
    if (poll_device || x){

        a+=answering_table[35]; // " present status is"
        if (poll_device) a+=answering_table[39]; // " device not found"
        else a+=answering_table[38]; // " device found"
        a+=answering_table[43]; // " sl"

        if (ap_mode_or_sta_mode){
            a+=answering_table[37]; // " in s t a mode"
            a+=answering_table[43]; // " sl"
            if (!is_station_mode_established) a+=answering_table[41]; // " offline"
            else a+=answering_table[40]; // " online"
            a+=answering_table[43]; // " sl"
        }
        else {
            a+=answering_table[36]; // " in a p mode"
            a+=answering_table[43]; // " sl"
        }
        if (poll_device) a+=answering_table[42]; // " please prepare well first"
        a+=answering_table[44]; // " sl sl"
    }
    if (x) mp3_response_local_custom=a;
    else if (poll_device){
        mp3_response_local_custom=a;
        return 1; // has some issue, need to report.
    }
    return 0; // no issue, even if need to report.
}


// ---------------------------------------------------------------------------------------- vars, main control/table processing.
cLinkedTables *voice_tables;
unsigned process_level=0, process_level_back=0;
void (*action_confirmed_callback)(unsigned)=0; // funcptr for callback after user confirmed, use parameter -1 to identify.
unsigned answering_table_idx=1; // set to 1 for startup.
String mp3_response_local_custom; // for customized words from local status.
String mp3_response_remote_custom; // for customized words from remote status.
unsigned flag_for_wdt_reset; // set to 1 to wdt reset in 5 seconds.
unsigned the_debug_data, the_debug_cnt;
unsigned the_voice_recog_enabled=1; // to control the recognition enabled or not.


// ---------------------------------------------------------------------------------------- table processing/common used.
mytable answering_table={ // ("/", single file), ("\", literal), ("|", add silence x 3).
    "",
    "|startup.mp3",                                                 // 1
    "hi, i m mary sl, the voice recognition device sl sl.",         // 2
    "the y 5 is established sl sl.",                                // 3
    "motor device is found, and is online sl sl.",                  // 4
    "excuse me sir sl, please say it again thank you sl sl.",       // 5
    "yes sir, what can i help you sl sl.",                          // 6
    "o fxxk sl you, ken woo sl, what is up sl sl.",                 // 7
    "excuse me, please say it again sl sl.",                        // 8

    // confirm for commit.
    "make sure: open the valve sl sl?",                             // 9
    "make sure: close the valve sl sl?",                            // 10
    "make sure: valve set auto sl sl?",                             // 11
    "make sure: lock valve sl sl?",                                 // 12
    "make sure: valve tiny open sl sl?",                            // 13
    "make sure: valve tiny close sl sl?",                           // 14
    "make sure: get online sl sl?",                                 // 15
    "make sure: get offline sl sl?",                                // 16
    "make sure: switch to a p mode sl sl?",                         // 17
    "make sure: switch to s t a mode sl sl?",                       // 18

    // results.
    "already in s t a mode sl sl",                                  // 19
    "already in a p mode sl sl",                                    // 20
    "present online sl sl",                                         // 21
    "present offline sl sl",                                        // 22
    "device found sl sl",                                           // 23
    "device disconnect sl sl",                                      // 24
    "i will switch to a p mode and reboot sl sl",                   // 25
    "i will switch to s t a mode and reboot sl sl",                 // 26
    "command done sl sl",                                           // 27
    "command failed sl sl",                                         // 28
    "/sample1.mp3",                                                 // 29
    "/sample2.mp3",                                                 // 30
    "\\19216811",                                                   // 31
    "action proceed doo doo sl sl",                                 // 32
    "action cancel bp sl sl",                                       // 33
    "yes sl",                                                       // 34

    // for composition.
    " present status is",                                           // 35
    " in a p mode",                                                 // 36
    " in s t a mode",                                               // 37
    " device found",                                                // 38
    " device not found",                                            // 39
    " online",                                                      // 40
    " offline",                                                     // 41
    " please set well",                                             // 42
    " sl",                                                          // 43
    " sl sl",                                                       // 44
    " doo",                                                         // 45
    " bp",                                                          // 46
    " oops sl",                                                     // 47

    "the speed is sl sl",                                           // 48
    "make sure: set auto-online sl sl?",                            // 49
    "make sure: disable auto-online sl sl?",                        // 50
    "make sure: disable voice recognition function sl sl?",         // 51
    "make sure: reboot sl sl?",                                     // 52
    "command done sl sl.",                                          // 53
    "please switch to s t a mode sl sl.",                           // 54
    "xx sl",                                                        // 55
    "xxx sl",                                                       // 56
};


// note: the tables structure, voc0(to recog) -> foo0(to invoke) -> voc1 -> foo1, etc.
void foo_level_common_forward(unsigned a){
    voice_tables->moveToNext(voice_tables);

    // get the action function ptr for later callback.
    action_confirmed_callback=(void (*)(unsigned))((*((char*(*)[])(voice_tables->getT())))[a]);

    voice_tables->moveToNext(voice_tables);
    LD3320::LD_ChangeTable(voice_tables->getT());

    process_level++;
    process_level_back=0;

    MY_DBG(Serial.printf("-------------------------> to level %u\r\n", process_level));
    MY_DBG(voice_tables->debug_dumpTable(*(char*(*)[])LD3320::LD_GetTable()));
}

void foo_level_common_backward(unsigned a){
    voice_tables->moveToPrev(voice_tables);
    voice_tables->moveToPrev(voice_tables);
    LD3320::LD_ChangeTable(voice_tables->getT());

    process_level--;
    process_level_back=0;

    MY_DBG(Serial.printf("-------------------------> to level %u\r\n", process_level));
    MY_DBG(voice_tables->debug_dumpTable(*(char*(*)[])LD3320::LD_GetTable()));
}

void foo_level_common_stay(unsigned a){
    MY_DBG(Serial.printf("-------------------------> stayed level %u\r\n", process_level));
    MY_DBG(voice_tables->debug_dumpTable(*(char*(*)[])LD3320::LD_GetTable()));

    process_level_back=0;

    answering_table_idx=8; // ask to repeat.
}

void foo_backward_common_yes(unsigned a){ // back to head.
    MY_DBG(Serial.printf("-------------------------> %u Yes\r\n", a));

    if (action_confirmed_callback) action_confirmed_callback(-1); // use -1 to identify been callback again.
    voice_tables=voice_tables->getFirst(); // turnaround to head.
    LD3320::LD_ChangeTable(voice_tables->getT()); // reload the new table.

    process_level=1;
    process_level_back=0;

    answering_table_idx=32; // action permit.
}

void foo_backward_common_no(unsigned a){ // back to head.
    MY_DBG(Serial.printf("-------------------------> %u No\r\n", a));

    voice_tables=voice_tables->getFirst(); // turnaround to head.
    LD3320::LD_ChangeTable(voice_tables->getT()); // reload the new table.

    process_level=1;
    process_level_back=0;

    answering_table_idx=33; // action cancelled.
}

void foo_backward_common_head(unsigned a){ // only back to head.
    voice_tables=voice_tables->getFirst(); // turnaround to head.
    LD3320::LD_ChangeTable(voice_tables->getT()); // reload the new table.

    process_level=1;
    process_level_back=0;

    MY_DBG(Serial.printf("-------------------------> %u HEAD\r\n", a));
}


// ---------------------------------------------------------------------------------------- tables of each level definitions.
// ============================================== level 0 (process_level==1).
mytable voc_tbl0={
    "xiao jie",             // 0 小姐
    "pu long gong",         // 1 普攏拱
    "woo own hon",          // 2
    "hoo own own",          // 3
    "ma li",                // 4 Mary
    "ma ri",                // 5
    "me li",                // 6
    "me ri",                // 7
    "xiao jie qi dong yu yin shi bie", // 8 小姐啟動語音識別

    "en zi si kou pu xu",   // 9 恩茲思扣撲虛
    "pu kou kou",           // 10
    "pu pu kou",            // 11
    "si zi",                // 12
    "zi si",                // 13
    "si zi gi",             // 14
    "sa je",                // 15
    "xiao zi",              // 16
    ""
};

void foo_level0_garbage(unsigned a){
    switch (a){
        case 9: MY_DBG(Serial.println("-------------------------> 9 en zi si kou pu xu\r\n")); break;
        case 10: MY_DBG(Serial.println("-------------------------> 10 pu kou kou\r\n")); break;
        case 11: MY_DBG(Serial.println("-------------------------> 11 pu pu kou\r\n")); break;
        case 12: MY_DBG(Serial.println("-------------------------> 12 si zi\r\n")); break;
        case 13: MY_DBG(Serial.println("-------------------------> 13 zi si\r\n")); break;
        case 14: MY_DBG(Serial.println("-------------------------> 14 si zi gi\r\n")); break;
        case 15: MY_DBG(Serial.println("-------------------------> 15 sa je\r\n")); break;
        case 16: MY_DBG(Serial.println("-------------------------> 16 xiao zi\r\n")); break;
        default: MY_DBG(Serial.println("-------------------------> xxxxxxxxx\r\n")); break;
    }
}

void foo_xiao_jie(unsigned a){
    MY_DBG(Serial.printf("-------------------------> %u 小姐\r\n", a));

    if (!the_voice_recog_enabled) return;

    foo_level_common_forward(a);

    answering_table_idx=6;
}
void foo_pu_long_gong(unsigned a){
    MY_DBG(Serial.printf("-------------------------> %u 普攏拱\r\n", a));

    if (!the_voice_recog_enabled) return;

    foo_level_common_forward(a);

    answering_table_idx=7;
}
void foo_ma_li(unsigned a){
    MY_DBG(Serial.printf("-------------------------> %u Mary\r\n", a));

    if (!the_voice_recog_enabled) return;

    foo_level_common_forward(a);

    answering_table_idx=34;
}
void foo_xiao_jie_qi_dong_yu_yin_shi_bie(unsigned a){
    MY_DBG(Serial.printf("-------------------------> %u 小姐啟動語音識別\r\n", a));

    the_voice_recog_enabled=1;

    foo_level_common_forward(a);

    answering_table_idx=7;
}

mytable foo_tbl0={
    (mycstr)foo_xiao_jie,                   // "xiao jie",             // 小姐
    (mycstr)foo_pu_long_gong,               // "pu long gong",         // 普攏拱
    (mycstr)foo_pu_long_gong,               // "woo own hon",
    (mycstr)foo_pu_long_gong,               // "hoo own own",
    (mycstr)foo_ma_li,                      // "ma li",                // Mary
    (mycstr)foo_ma_li,                      //
    (mycstr)foo_ma_li,                      //
    (mycstr)foo_ma_li,                      //
    (mycstr)foo_xiao_jie_qi_dong_yu_yin_shi_bie, //

    (mycstr)foo_level0_garbage,             // "en zi si kou pu xu",   // 恩茲思扣撲虛
    (mycstr)foo_level0_garbage,             // "pu kou kou",
    (mycstr)foo_level0_garbage,             // "pu pu kou",
    (mycstr)foo_level0_garbage,             // "si zi",
    (mycstr)foo_level0_garbage,             // "zi si",
    (mycstr)foo_level0_garbage,             // "si zi gi",
    (mycstr)foo_level0_garbage,             // "sa je",
    (mycstr)foo_level0_garbage,             // "xiao zi",
    ""
};


// ============================================== level 1 (process_level==2).
mytable voc_tbl1={
    "fa men kai qi",        // 0 閥門開啟
    "kai qi fa men",        // 1 開啟閥門
    "fa men guan bi",       // 2 閥門關閉
    "guan bi fa men",       // 3 關閉閥門
    "xian zai zhuai su",    // 4 現在轉速
    "fa men zi dong",       // 5 閥門自動
    "fa men suo ding",      // 6 閥門鎖定
    "fa men wei kai",       // 7 閥門微開
    "fa men wei guan",      // 8 閥門微關

    "lian xian zhuang tai", // 9 "連線狀態"
    "wai pi mu shi",        // 10 "歪壞模式"
    "qing shang xian",      // 11 "請上線"
    "qing li xian",         // 12 "請離線"
    "qie huan cheng ai pi mu shi", // 13 "切換成 a p 模式"
    "qie huan cheng shi de xun mu shi", // 14 "切換成 station 模式"
    "qie huan cheng ai shi ti ai mu shi", // 15 "切換成 station 模式"
    "chang ge",             // 16 "唱歌"
    "bo ge",                // 17 "唱歌"
    "fang ge",              // 18 "唱歌"
    "ma da ai pi",          // 19 "馬達 i p"
    "ben de ai pi",         // 20 "本地 i p"
    "ben he ai pi",         // 21 "本地 i p"
    "ben se ai pi",         // 22 "本地 i p"
    "ben te ai pi",         // 23 "本地 i p"
    "wai fi ai pi",         // 24 "歪壞 i p"
    "wa fi ai pi",          // 25 "歪壞 i p"
    "qi dong kai ji lian wang", // 26 "啟動開機連網"
    "qu xiao kai ji lian wang", // 27 "取消開機連網"
    "tui chu yu yin shi bie", // 28 "退出語音識別"
    "chong zhi",            // 29 "重置"
    "zeng jia yin liang",   // 30 "增加音量"
    "jian shao yin liang",  // 31 "減少音量"
    "zeng jia ling min duo", // 32 "增加零敏度"
    "jian shao ling min duo", // 33 "減少零敏度"
    "hui fu ling min duo",  // 34 "恢復零敏度"
    "huei fu ling min duo", // 35 "恢復零敏度"
    "ling min duo zui gao", // 36 "零敏度最高"
    "ling min duo zui di", // 37 "零敏度最低"

    "chong xin kai ji", // 38 "重新開機"
    "chong xin qi dong", // 39 "重新啟動"

    "en zi si kou pu xu",   // 40 恩茲思扣撲虛
    "pu kou kou",           // 41
    "pu pu kou",            // 42
    "si zi",                // 43
    "zi si",                // 44
    "si zi gi",             // 45
    "sa je",                // 46
    "xiao zi",              // 47
    ""
};

void foo_level1_garbage(unsigned a){
    switch (a) {
        case 40: MY_DBG(Serial.println("-------------------------> 40 en zi si kou pu xu\r\n")); break;
        case 41: MY_DBG(Serial.println("-------------------------> 41 pu kou kou\r\n")); break;
        case 42: MY_DBG(Serial.println("-------------------------> 42 pu pu kou\r\n")); break;
        case 43: MY_DBG(Serial.println("-------------------------> 43 si zi\r\n")); break;
        case 44: MY_DBG(Serial.println("-------------------------> 44 zi si\r\n")); break;
        case 45: MY_DBG(Serial.println("-------------------------> 45 si zi gi\r\n")); break;
        case 46: MY_DBG(Serial.println("-------------------------> 46 sa je\r\n")); break;
        case 47: MY_DBG(Serial.println("-------------------------> 47 xiao zi\r\n")); break;
        default: MY_DBG(Serial.println("-------------------------> xxxxxxxxx\r\n")); break;
    }

    foo_level_common_stay(a);
}

void foo_fa_men_kai_qi(unsigned a){
    MY_DBG(Serial.println("-------------------------> 閥門開啟\r\n"));

    if (a==-1){ // being callbacked.
        successive_cmd_issue=2;
        return;
    }

    foo_level_common_forward(a);

    answering_table_idx=9;
}
void foo_fa_men_guan_bi(unsigned a){
    MY_DBG(Serial.println("-------------------------> 閥門關閉\r\n"));

    if (a==-1){ // being callbacked.
        successive_cmd_issue=3;
        return;
    }

    foo_level_common_forward(a);

    answering_table_idx=10;
}
void foo_xian_zai_zhuai_su(unsigned a){
    successive_cmd_issue=4;
    mp3_response_remote_custom="\\"; // means it is non-null and will be added the remote response in main-loop and literally.

    foo_backward_common_head(a);

    answering_table_idx=48;

    MY_DBG(Serial.println("-------------------------> 現在轉速\r\n"));
}
void foo_fa_men_zi_dong(unsigned a){
    MY_DBG(Serial.println("-------------------------> 閥門自動\r\n"));

    if (a==-1){ // being callbacked.
        successive_cmd_issue=6;
        return;
    }

    foo_level_common_forward(a);

    answering_table_idx=11;
}
void foo_fa_men_suo_ding(unsigned a){
    MY_DBG(Serial.println("-------------------------> 閥門鎖定\r\n"));

    if (a==-1){ // being callbacked.
        successive_cmd_issue=1;
        return;
    }

    foo_level_common_forward(a);

    answering_table_idx=12;
}
void foo_fa_men_wei_kai(unsigned a){
    MY_DBG(Serial.println("-------------------------> 閥門微開\r\n"));

    if (a==-1){ // being callbacked.
        successive_cmd_issue=7;
        return;
    }

    foo_level_common_forward(a);

    answering_table_idx=13;
}
void foo_fa_men_wei_guan(unsigned a){
    MY_DBG(Serial.println("-------------------------> 閥門微關\r\n"));

    if (a==-1){ // being callbacked.
        successive_cmd_issue=11;
        return;
    }

    foo_level_common_forward(a);

    answering_table_idx=14;
}
void foo_lian_xian_zhuang_tai(unsigned a){
    foo_backward_common_head(a);

    RemoteAccsAux(1);
    answering_table_idx=-1;

    MY_DBG(Serial.println("-------------------------> 連線狀態\r\n"));
}
void foo_wai_pi_mu_shi(unsigned a){
    foo_backward_common_head(a);

    if (ap_mode_or_sta_mode) answering_table_idx=19;
    else answering_table_idx=20;

    MY_DBG(Serial.println("-------------------------> 歪壞模式\r\n"));
}
void foo_qing_shang_xian(unsigned a){
    MY_DBG(Serial.println("-------------------------> 請上線\r\n"));

    if (a==-1){ // being callbacked.
        if (ap_mode_or_sta_mode){ // sta mode.
            successive_cmd_status=1;
            sta_do_the_connect=1; // demand to connect wifi.
        }
        else successive_cmd_issue=-1; // set as -1 for local treatment which mainly check the network status.
        return;
    }

    foo_level_common_forward(a);

    answering_table_idx=15;
}
void foo_qing_li_xian(unsigned a){
    MY_DBG(Serial.println("-------------------------> 請離線\r\n"));

    if (a==-1){ // being callbacked.

        successive_cmd_status=10; // special use the 10 for code flow reset. ////////////
        sta_do_the_connect=0; // no, must use this line otherwise successive_cmd_status may not reached.

        mp3_response_local_custom=answering_table[22]; // directly response; must set successive_cmd_issue=-1.
        successive_cmd_issue=-1; // set as -1 for local treatment.
        return;
    }

    foo_level_common_forward(a);

    answering_table_idx=16;
}
void foo_qie_huan_cheng_ai_pi_mu_shi(unsigned a){
    MY_DBG(Serial.println("-------------------------> 切換成 a p 模式\r\n"));

    if (a==-1){ // being callbacked.
        if (ap_mode_or_sta_mode){ // is in sta mode.
            sta_do_the_connect=0;
            ap_mode_or_sta_mode=0;

            flag_for_wdt_reset=1;

            #if MYDB
                cDataBase("/settings").ReplaceWordsEx("ap_mode_or_sta_mode", 0, ap_mode_or_sta_mode);
                cDataBase("/settings").ReplaceWordsEx("sta_do_the_connect", 0, 0);
            #endif // MYDB

            mp3_response_local_custom=answering_table[25]; // directly response.
            successive_cmd_issue=-1; // for local treatment.
        }
        else {
            mp3_response_local_custom=answering_table[20]; // directly response.
            successive_cmd_issue=-1; // for local treatment.
        }
        return;
    }

    foo_level_common_forward(a);

    answering_table_idx=17;
}
void foo_qie_huan_cheng_shi_de_xun_mu_shi(unsigned a){
    MY_DBG(Serial.println("-------------------------> 切換成 station 模式\r\n"));

    if (a==-1){ // being callbacked.
        if (!ap_mode_or_sta_mode){ // is in ap mode.
            ap_mode_or_sta_mode=1;

            flag_for_wdt_reset=1;

            #if MYDB
                cDataBase("/settings").ReplaceWordsEx("ap_mode_or_sta_mode", 0, ap_mode_or_sta_mode);
            #endif // MYDB

            mp3_response_local_custom=answering_table[26]; // directly response.
            successive_cmd_issue=-1; // for local treatment.
        }
        else {
            mp3_response_local_custom=answering_table[19]; // directly response.
            successive_cmd_issue=-1; // for local treatment.
        }
        return;
    }

    foo_level_common_forward(a);

    answering_table_idx=18;
}
void foo_chang_ge(unsigned a){
    MY_DBG(Serial.println("-------------------------> 唱歌\r\n"));

    foo_backward_common_head(a);

    answering_table_idx=29;
}
void foo_ma_da_ai_pi(unsigned a){
    MY_DBG(Serial.println("-------------------------> 馬達 i p\r\n"));

    if (is_station_mode_established && target_dev_ip!=""){
        mp3_response_local_custom="\\"+target_dev_ip; // directly response.
        successive_cmd_issue=-1; // set as -1 for local treatment.
    }
    else successive_cmd_issue=-1; // set as -1 for local treatment.

    foo_backward_common_head(a);
}
void foo_ben_de_ai_pi(unsigned a){
    MY_DBG(Serial.println("-------------------------> 本地 i p\r\n"));

    if (is_station_mode_established && my_local_ip!=""){
        mp3_response_local_custom="\\"+my_local_ip; // directly response.
        successive_cmd_issue=-1; // set as -1 for local treatment.
    }
    else successive_cmd_issue=-1; // set as -1 for local treatment.

    foo_backward_common_head(a);
}
void foo_wai_fi_ai_pi(unsigned a){
    MY_DBG(Serial.println("-------------------------> 歪壞 i p\r\n"));

    if (is_station_mode_established && *remote_ap_ip){
        mp3_response_local_custom=String("\\")+remote_ap_ip; // directly response.
        successive_cmd_issue=-1; // set as -1 for local treatment.
    }
    else successive_cmd_issue=-1; // set as -1 for local treatment.

    foo_backward_common_head(a);
}
void qi_dong_kai_ji_lian_wang(unsigned a){ // 26 "啟動開機連網"
    MY_DBG(Serial.println("-------------------------> 啟動開機連網\r\n"));

    if (a==-1){ // being callbacked.
        if (ap_mode_or_sta_mode){ // is in sta mode.

            #if MYDB
                cDataBase("/settings").ReplaceWordsEx("ap_mode_or_sta_mode", 0, ap_mode_or_sta_mode);
                cDataBase("/settings").ReplaceWordsEx("sta_do_the_connect", 0, 1);
            #endif // MYDB

            mp3_response_local_custom=answering_table[53]; // directly response.
            successive_cmd_issue=-1; // for local treatment.
        }
        else {
            mp3_response_local_custom=answering_table[54]; // directly response.
            successive_cmd_issue=-1; // for local treatment.
        }
        return;
    }

    foo_level_common_forward(a);

    answering_table_idx=49;
}
void qu_xiao_kai_ji_lian_wang(unsigned a){ // 27 "取消開機連網"
    MY_DBG(Serial.println("-------------------------> 取消開機連網\r\n"));

    if (a==-1){ // being callbacked.

        #if MYDB
            cDataBase("/settings").ReplaceWordsEx("sta_do_the_connect", 0, 0);
        #endif // MYDB

        mp3_response_local_custom=answering_table[53]; // directly response.
        successive_cmd_issue=-1; // for local treatment.

        return;
    }

    foo_level_common_forward(a);

    answering_table_idx=50;
}
void tui_chu_yu_yin_shi_bie(unsigned a){ // 28 "退出語音識別"
    MY_DBG(Serial.println("-------------------------> 退出語音識別\r\n"));

    if (a==-1){ // being callbacked.

        //// LD3320Stop(); // not work here, run it elsewhere.
        the_voice_recog_enabled=0; // disabled it.

        mp3_response_local_custom=answering_table[53]; // directly response.
        successive_cmd_issue=-1; // for local treatment.

        return;
    }

    foo_level_common_forward(a);

    answering_table_idx=51;
}
void chong_zhi(unsigned a){ // 29 "重置"
    MY_DBG(Serial.println("-------------------------> 重置\r\n"));

    if (a==-1){ // being callbacked.

        flag_for_wdt_reset=1;

        mp3_response_local_custom=answering_table[53]; // directly response.
        successive_cmd_issue=-1; // for local treatment.

        return;
    }

    foo_level_common_forward(a);

    answering_table_idx=52;
}
void foo_zeng_jia_yin_liang(unsigned a){ // 30 "增加音量"
    foo_backward_common_head(a);

    if (set_volume+2<15) set_volume+=2; // adjust volume.
    else set_volume=15;

    #if MYDB
        cDataBase("/settings").ReplaceWordsEx("set_volume", 0, set_volume);
    #endif // MYDB

    if (set_volume==15) answering_table_idx=1; else answering_table_idx=2;

    MY_DBG(Serial.println("-------------------------> 增加音量\r\n"));
}
void foo_jian_shao_yin_liang(unsigned a){ // 31 "減少音量"
    foo_backward_common_head(a);

    if ((int)set_volume-2>0) set_volume-=2; // adjust volume.
    else set_volume=1; // mute is worse.

    #if MYDB
        cDataBase("/settings").ReplaceWordsEx("set_volume", 0, set_volume);
    #endif // MYDB

    if (set_volume==1) answering_table_idx=1; else answering_table_idx=2;

    MY_DBG(Serial.println("-------------------------> 減少音量\r\n"));
}
void foo_zeng_jia_ling_min_duo(unsigned a){ // 32 "增加零敏度"
    MY_DBG(Serial.println("-------------------------> 增加零敏度\r\n"));

    asr_sensitive_l++;
    foo_backward_common_head(a);

    #if MYDB
        cDataBase("/settings").ReplaceWordsEx("asr_sensitive_l", 0, asr_sensitive_l);
    #endif // MYDB

    mp3_response_local_custom=answering_table[53]; // directly response.
    successive_cmd_issue=-1; // for local treatment.
}
void foo_jian_shao_ling_min_duo(unsigned a){ // 33 "減少零敏度"
    MY_DBG(Serial.println("-------------------------> 減少零敏度\r\n"));

    asr_sensitive_l--;
    foo_backward_common_head(a);

    #if MYDB
        cDataBase("/settings").ReplaceWordsEx("asr_sensitive_l", 0, asr_sensitive_l);
    #endif // MYDB

    mp3_response_local_custom=answering_table[53]; // directly response.
    successive_cmd_issue=-1; // for local treatment.
}
void foo_hui_fu_ling_min_duo(unsigned a){ // 33 "恢復零敏度"
    MY_DBG(Serial.println("-------------------------> 恢復零敏度\r\n"));

    asr_sensitive_l=3;
    foo_backward_common_head(a);

    #if MYDB
        cDataBase("/settings").ReplaceWordsEx("asr_sensitive_l", 0, asr_sensitive_l);
    #endif // MYDB

    mp3_response_local_custom=answering_table[53]; // directly response.
    successive_cmd_issue=-1; // for local treatment.
}
void foo_ling_min_duo_zui_gao(unsigned a){ // 33 "零敏度最高"
    MY_DBG(Serial.println("-------------------------> 零敏度最高\r\n"));

    asr_sensitive_l=6;
    foo_backward_common_head(a);

    #if MYDB
        cDataBase("/settings").ReplaceWordsEx("asr_sensitive_l", 0, asr_sensitive_l);
    #endif // MYDB

    mp3_response_local_custom=answering_table[53]; // directly response.
    successive_cmd_issue=-1; // for local treatment.
}
void foo_ling_min_duo_zui_di(unsigned a){ // 33 "零敏度最低"
    MY_DBG(Serial.println("-------------------------> 零敏度最低\r\n"));

    asr_sensitive_l=0;
    foo_backward_common_head(a);

    #if MYDB
        cDataBase("/settings").ReplaceWordsEx("asr_sensitive_l", 0, asr_sensitive_l);
    #endif // MYDB

    mp3_response_local_custom=answering_table[53]; // directly response.
    successive_cmd_issue=-1; // for local treatment.
}

mytable foo_tbl1={
    (mycstr)foo_fa_men_kai_qi,              // "fa men kai qi",        // 閥門開啟
    (mycstr)foo_fa_men_kai_qi,              // "kai qi fa men",        // 開啟閥門
    (mycstr)foo_fa_men_guan_bi,             // "fa men guan bi",       // 閥門關閉
    (mycstr)foo_fa_men_guan_bi,             // "guan bi fa men",       // 關閉閥門
    (mycstr)foo_xian_zai_zhuai_su,          // "xian zai zhuai su",    // 現在轉速
    (mycstr)foo_fa_men_zi_dong,             // "fa men zi dong",       // 閥門自動
    (mycstr)foo_fa_men_suo_ding,            // "fa men suo ding",      // 閥門鎖定
    (mycstr)foo_fa_men_wei_kai,             // "fa men wei kai",       // 閥門微開
    (mycstr)foo_fa_men_wei_guan,            // "fa men wei guan",      // 閥門微關

    (mycstr)foo_lian_xian_zhuang_tai,
    (mycstr)foo_wai_pi_mu_shi,
    (mycstr)foo_qing_shang_xian,
    (mycstr)foo_qing_li_xian,
    (mycstr)foo_qie_huan_cheng_ai_pi_mu_shi,
    (mycstr)foo_qie_huan_cheng_shi_de_xun_mu_shi,
    (mycstr)foo_qie_huan_cheng_shi_de_xun_mu_shi,
    (mycstr)foo_chang_ge,
    (mycstr)foo_chang_ge,
    (mycstr)foo_chang_ge,
    (mycstr)foo_ma_da_ai_pi,
    (mycstr)foo_ben_de_ai_pi,
    (mycstr)foo_ben_de_ai_pi,
    (mycstr)foo_ben_de_ai_pi,
    (mycstr)foo_ben_de_ai_pi,
    (mycstr)foo_wai_fi_ai_pi,
    (mycstr)foo_wai_fi_ai_pi,

    (mycstr)qi_dong_kai_ji_lian_wang,
    (mycstr)qu_xiao_kai_ji_lian_wang,
    (mycstr)tui_chu_yu_yin_shi_bie,
    (mycstr)chong_zhi,

    (mycstr)foo_zeng_jia_yin_liang,
    (mycstr)foo_jian_shao_yin_liang,
    (mycstr)foo_zeng_jia_ling_min_duo,
    (mycstr)foo_jian_shao_ling_min_duo,
    (mycstr)foo_hui_fu_ling_min_duo,
    (mycstr)foo_hui_fu_ling_min_duo,
    (mycstr)foo_ling_min_duo_zui_gao,
    (mycstr)foo_ling_min_duo_zui_di,

    (mycstr)chong_zhi, // 29 "重置" "chong xin kai ji", // 38 "重新開機"
    (mycstr)chong_zhi, // 29 "重置" "chong xin qi dong", // 39 "重新啟動"

    (mycstr)foo_level1_garbage,
    (mycstr)foo_level1_garbage,
    (mycstr)foo_level1_garbage,
    (mycstr)foo_level1_garbage,
    (mycstr)foo_level1_garbage,
    (mycstr)foo_level1_garbage,
    (mycstr)foo_level1_garbage,
    (mycstr)foo_level1_garbage,
    ""
};


// ============================================== level 2 (process_level==3).
mytable voc_tbl2={
    "dui",                  // 0 "對"
    "shi",                  // 1 "是"
    "shi di",               // 2 "是的"
    "yes",                  // 3 "yes"
    "ye shi",               // 4 "yes"
    "bu dui",               // 5 "不對"
    "bu shi",               // 6 "不是"
    "cu",                   // 7 "錯"

    "shi de",               // 8 "是的" // false positive
    "no",                   // 9 // noise
    "en zi si kou pu xu",   // 10 恩茲思扣撲虛
    "pu kou kou",           // 11
    "pu pu kou",            // 12
    "si zi",                // 13
    "zi si",                // 14
    "si zi gi",             // 15
    "sa je",                // 16
    "xiao zi",              // 17
    ""
};

mytable foo_tbl2={
    (mycstr)foo_backward_common_yes,
    (mycstr)foo_backward_common_yes,
    (mycstr)foo_backward_common_yes,
    (mycstr)foo_backward_common_yes,
    (mycstr)foo_backward_common_yes,
    (mycstr)foo_backward_common_no,
    (mycstr)foo_backward_common_no,
    (mycstr)foo_backward_common_no,

    (mycstr)foo_level_common_stay,
    (mycstr)foo_level_common_stay,
    (mycstr)foo_level_common_stay,
    (mycstr)foo_level_common_stay,
    (mycstr)foo_level_common_stay,
    (mycstr)foo_level_common_stay,
    (mycstr)foo_level_common_stay,
    (mycstr)foo_level_common_stay,
    (mycstr)foo_level_common_stay,
    (mycstr)foo_level_common_stay,
    ""
};




// --------------------------------------------------- main -------------------------------------------------------------
void setup(){
    MY_DBG(Serial.begin(115200));
    MY_DBG(while (!Serial) delay(10));
    MY_DBG(delay(3000));
    MY_DBG(Serial.printf("\r\nStart\r\n"));

    // IoTs
    the_iot_setup();

    // LD3320
    LD3320::LD_Init(); // init.
    nAsrStatus=LD_ASR_NONE; // means regularly doing the asr or mp3s time by time.

    // prepare tables
    voice_tables=new cLinkedTables;
    voice_tables->assignTable(voc_tbl0);
    voice_tables->addTable(foo_tbl0);
    voice_tables->addTable(voc_tbl1);
    voice_tables->addTable(foo_tbl1);
    voice_tables->addTable(voc_tbl2);
    voice_tables->addTable(foo_tbl2);

    LD3320::LD_ChangeTable(voice_tables->getT());

    process_level=1;
    if (sta_do_the_connect) successive_cmd_status=1;
}

void loop(){
    #if MP3_IN_ISR
        halt_wdt();
    #endif


    // for auto responder, or enquiry.
    if (!answering_table_idx){

        // the order of the following lines here is of concern.
        // until there is no response, and still has successive_cmd_issue pending,
        // which means something not available or is intented to,
        // so we check and add new response.
        // also if needs remote result, we handle it here.
        //
        // as the overall effects: significant answering_table_idx value to access first,
        // second, check mp3_response_local_custom or RemoteAccsAux which has something needs to response,
        // third, check mp3_response_remote_custom has something to response.
        if (successive_cmd_issue){ // means might have something wrong.
            if (mp3_response_local_custom!="" || RemoteAccsAux()){ // either directly response or need to check connection states.
                successive_cmd_issue=0;
                mp3_response_remote_custom="";

                // this new state answering_table_idx=-1, has been avoided from an older non-zero value of answering_table_idx,
                // which needed to resolve and play, and then the -1 yields play for the second time.
                // so this new state will correspond to mp3_response_local_custom to play.
                answering_table_idx=-1;
            }
        }
        else if (mp3_response_remote_custom!=""){ // remote command was success and got the result. should set non-null elsewhere.
            mp3_response_local_custom=mp3_response_remote_custom; // note, mp3_response_remote_custom was predefined into non-null.
            mp3_response_local_custom+=target_dev_last_sent;
            mp3_response_remote_custom="";
            answering_table_idx=-1; // set remote result to mp3_response_local_custom to play.
        }

        switch (successive_cmd_status){
            case 2: // online
                answering_table_idx=3; // online.
                successive_cmd_status++;
                break;
            case 4: // found the device
                answering_table_idx=4; // got the device.
                successive_cmd_status++;
                break;
            default: break;
        }
    }
    else if (!LD3320MP3::isOpened()){ // play the auto-response sound.
        LD3320::StopASR(); // since it is modified to idle.
        nAsrStatus=LD_ASR_NONE; // since it is modified to idle.

        if (answering_table_idx==-1){ // designate the customized words provided by mp3_response_local_custom to play.
            MY_DBG(Serial.println(mp3_response_local_custom));
            if (*(mp3_response_local_custom.c_str())=='\\'){
                MP3Sentence t;
                t.Literal(mp3_response_local_custom.c_str(), 2);
            }
            else MP3Sentence t(mp3_response_local_custom.c_str(), 2);
            mp3_response_local_custom="";
        }
        else if (*answering_table[answering_table_idx]=='/') // a single mp3 file.
            LD3320MP3 t(answering_table[answering_table_idx]);
        else if (*answering_table[answering_table_idx]=='\\'){ // speak literally.
            MY_DBG(Serial.println(answering_table[answering_table_idx]));
            MP3Sentence t;
            t.Literal(answering_table[answering_table_idx], 2);
        }
        else if (*answering_table[answering_table_idx]=='|'){ // multiple files and add 3 silence.
            MY_DBG(Serial.println(answering_table[answering_table_idx]));
            MP3Sentence t;
            t.AddSilence(answering_table[answering_table_idx], 3);
        }
        else {
            MY_DBG(Serial.println(answering_table[answering_table_idx]));
            MP3Sentence t(answering_table[answering_table_idx]);
        }

        switch (answering_table_idx){
            case 1: answering_table_idx=2; break;
            default: answering_table_idx=0; break;
        }
    }


    static unsigned tmo=micros();
    if (Timeout(1350, tmo)){
        MY_DBG(Serial.printf("%u %u %u\r\n", nAsrStatus, successive_cmd_issue, the_debug_cnt));

        // back to level 0 after about 10 seconds.
        if ((process_level>=2) && ++process_level_back>8){
            /// here, must like that or asr may mulfunction.
            ///LD3320::StopASR(); // it is during sampling so we have to stop it or fault.
            ///nAsrStatus=LD_ASR_NONE; // since stop would back to idle.
            ///LD3320::LD_Reset(); nAsrStatus=LD_ASR_NONE; // must this line or asr may mulfunction.

            foo_backward_common_head(0);
            answering_table_idx=56; // remind the end or nothing-happen message.
        }

        if (flag_for_wdt_reset && ++flag_for_wdt_reset>5) while (1);
    }

    switch (nAsrStatus){
        case LD_ASR_IDLE:
        case LD_ASR_BUSY:
        case LD_MP3_DATA: break;
        case LD_ASR_NONE: // regularly run the asr or play an mp3 file.
            if (LD3320::RunMP3()) break; // use LD3320MP3 tmp("sample.mp3") first to trigger.
            else if (!LD3320::RunASR()){
                MY_DBG(Serial.println("ASR fault and reset and retry."));
                LD3320::LD_Reset();
                nAsrStatus=LD_ASR_NONE;
                delay(50);
            }
            break;
        case LD_ASR_FOUNDOK: // asr success.
            #if 1
            {
                unsigned b=LD3320::LD_GetResult();
                ((void(*)(unsigned))((*(voice_tables->nxtT()))[b]))(b); // call the target.
            }
            #endif
            nAsrStatus=LD_ASR_NONE;
            break;
        case LD_ASR_FOUNDZERO:
            MY_DBG(Serial.println("ASR done but no result."));
            nAsrStatus=LD_ASR_NONE;
            if (!LD3320MP3::isOpened() && (process_level>=2)) answering_table_idx=47; // not the case which asr been stopped.
            break;
        case LD_MP3_INTR: // set by mp3 intr.
            #if !(MP3_IN_ISR)
                LD3320::ProcessIntMP3();
            #endif
            break;
        case LD_MP3_REST:
            LD3320::LD_Reset();
            nAsrStatus=LD_ASR_NONE;
            break;
        default: // LD_ASR_ERROR from busy; LD_ASR_EXCEPTION or LD_ASR_INTERR from intr.
            MY_DBG(Serial.printf("Unexpected error (%u).\r\n", nAsrStatus));
            LD3320::LD_Reset();
            nAsrStatus=LD_ASR_NONE;
            delay(50);
            break;
    }

    loop_IoT();
}
